import asyncio
import functools
import json
import logging
import random
import ssl
import string
import sys
from asyncio import Future, AbstractEventLoop
from asyncio import TimeoutError
from datetime import datetime
from enum import Enum
from hashlib import md5
from time import time
from typing import Optional, List, TypeVar, Iterable, Callable, Awaitable, Tuple, Union, Any

import paho.mqtt.client as mqtt
from aiohttp import ClientSession

from meross_iot.controller.device import BaseDevice, HubDevice, GenericSubDevice
from meross_iot.device_factory import (
    build_meross_device_from_abilities,
    build_meross_subdevice,
    build_meross_device_from_known_types,
)
from meross_iot.error_budget import ErrorBudgetManager
from meross_iot.http_api import MerossHttpClient
from meross_iot.model.constants import DEFAULT_COMMAND_TIMEOUT, DEFAULT_MQTT_PORT
from meross_iot.model.enums import Namespace, OnlineStatus
from meross_iot.model.exception import (
    CommandTimeoutError,
    CommandError,
    UnknownDeviceType
)
from meross_iot.model.http.device import HttpDeviceInfo
from meross_iot.model.http.subdevice import HttpSubdeviceInfo
from meross_iot.model.push.factory import parse_push_notification
from meross_iot.model.push.generic import GenericPushNotification
from meross_iot.model.push.online import OnlinePushNotification
from meross_iot.model.push.unbind import UnbindPushNotification
from meross_iot.utilities.mqtt import (
    generate_mqtt_password,
    generate_client_and_app_id,
    build_client_response_topic,
    build_client_user_topic,
    verify_message_signature,
    device_uuid_from_push_notification,
    build_device_request_topic,
)
from meross_iot.utilities.network import extract_domain

logging.basicConfig(
    format="%(levelname)s:%(message)s", level=logging.INFO, stream=sys.stdout
)
_LOGGER = logging.getLogger(__name__)

_CONNECTION_DROP_UPDATE_SCHEDULE_INTERVAL = 2

T = TypeVar("T", bound=BaseDevice)  # Declare type variable
ManagerPushNotificationHandlerType = Callable[[GenericPushNotification, List[BaseDevice], 'MerossManager'], Awaitable]

_PENDING_FUTURES = []

_DEFAULT_HEADERS = {"Content-Type": "application/json"}


def _mqtt_key_from_domain_port(domain: str, port: int) -> str:
    return f"{domain}:{port}"


class TransportMode(Enum):
    MQTT_ONLY = 0
    LAN_HTTP_FIRST = 1
    LAN_HTTP_FIRST_ONLY_GET = 2


class MqttConnectionStatus(Enum):
    DISCONNECTED = "DISCONNECTED"
    CONNECTING = "CONNECTING"
    CONNECTED = "CONNECTED"


class MerossManager(object):
    """
    This class implements a full-features Meross Client, which provides device discovery and registry.
    *Note*: The manager must be initialized before invoking any of its discovery/registry methods.
    """

    def __init__(
            self,
            http_client: MerossHttpClient,
            auto_reconnect: Optional[bool] = True,
            mqtt_skip_cert_validation: bool = False,
            ca_cert: Optional[str] = None,
            loop: Optional[AbstractEventLoop] = None,
            mqtt_override_server: Optional[Tuple[str, int]] = None,
            auto_discovery_on_connection: bool = True,
            *args,
            **kwords,
    ) -> None:
        """
        Constructor
        :param http_client: MerossHttpClient object that is used by the manager to query the HTTP API
                            (for device discovery, etc)
        :param auto_reconnect: (Optional) When True, the mqtt client will automatically reconnect when the connection
                               drops. Defaults to True.
        :param mqtt_skip_cert_validation: (Optional) When set the Manager will accept unverified SSL/TLS certificates
                                          from the remote MQTT server. Defaults to False.
        :param ca_cert: (Optional) Path to the PEM certificate to trust (Intermediate/CA)
        :param loop: (Optional) Asyncio loop to use
        :param args:
        :param mqtt_override_server: (Optional) Tuple (hostname, port) of the MQTT server to use for MQTT connection.
                                     When None (default), the hostname will be extracted by the domain/reserved domain
                                     obtained via HTTP API, and port 443 will be used.
        :param auto_discovery_on_connection: (Optional) When set instructs the manager to issue a discovery as soon as
                                             the mqtt connection is established against the MQTT broker (defaults to True)
        """

        # Store local attributes
        self._http_client = http_client
        self._cloud_creds = self._http_client.cloud_credentials
        self._auto_reconnect = auto_reconnect
        self._ca_cert = ca_cert
        self._app_id, self._client_id = generate_client_and_app_id()
        self._pending_messages_futures = {}
        self._device_registry = DeviceRegistry()
        self._push_coros = []
        self._mqtt_skip_validation = mqtt_skip_cert_validation
        self._mqtt_clients = {}
        self._mqtt_connected_and_subscribed = {}
        self._auto_discovery_on_connection = auto_discovery_on_connection

        # By default, assume MQTT-Only transport mode
        self._default_transport_mode = TransportMode.MQTT_ONLY
        self._error_budget_manager = ErrorBudgetManager()

        # Default proxy setup
        self._enable_proxy = False
        self._proxy_type = None
        self._proxy_addr = None
        self._proxy_port = None

        # Setup synchronization primitives
        self._mqtt_looper_task = None
        self._loop = asyncio.get_event_loop() if loop is None else loop

        # Prepare MQTT info
        self._mqtt_password = generate_mqtt_password(user_id=self._cloud_creds.user_id, key=self._cloud_creds.key)
        self._client_response_topic = build_client_response_topic(
            user_id=self._cloud_creds.user_id, app_id=self._app_id
        )
        self._user_topic = build_client_user_topic(user_id=self._cloud_creds.user_id)
        self._override_mqtt_server = mqtt_override_server

    @property
    def auto_discovery_on_connection(self) -> bool:
        """When set, tells the manager to automatically issue a discovery after a successful connection"""
        return self._auto_discovery_on_connection

    @auto_discovery_on_connection.setter
    def auto_discovery_on_connection(self, val: bool):
        self._auto_discovery_on_connection = val

    @property
    def default_transport_mode(self) -> TransportMode:
        return self._default_transport_mode

    @default_transport_mode.setter
    def default_transport_mode(self, value: TransportMode) -> None:
        self._default_transport_mode = value

    def _get_client_from_domain_port(self, client: mqtt.Client) -> Tuple[Optional[str], Optional[int]]:
        for k, v in self._mqtt_clients.items():
            if v == client:
                domain, port = k.split(":")
                return domain, int(port)
        return None, None

    async def _async_get_create_mqtt_client(self, domain: str, port: int) -> mqtt.Client:
        """
        Retrieves the mqtt_client for the given domain/port combination.
        If not existing, a new one is created
        """
        dict_key = _mqtt_key_from_domain_port(domain=domain, port=port)
        client = self._mqtt_clients.get(dict_key)
        if client is not None:
            _LOGGER.debug("MQTT Client for %s already available.", dict_key)
        else:
            _LOGGER.info("Allocating new mqtt client for %s...", dict_key)
            client = self._new_mqtt_client()
            client.user_data_set(dict_key)
            if self._enable_proxy:
                _LOGGER.info("Proxy configuration set for newly created client")
                client.proxy_set(proxy_type=self._proxy_type, proxy_addr=self._proxy_addr, proxy_port=self._proxy_port)
            self._mqtt_clients[dict_key] = client
        # Init the client
        if not client.is_connected():
            conn_evt = self._mqtt_connected_and_subscribed.get(dict_key)  # type: asyncio.Event
            if conn_evt is None:
                conn_evt = asyncio.Event()
                _LOGGER.debug("MQTT client connecting to %s:%d", domain, port)
                client.connect(host=domain, port=port, keepalive=30)
                self._mqtt_connected_and_subscribed[dict_key] = conn_evt
            # Start the client looper
            client.loop_start()
            # Wait for the client to connect
            await conn_evt.wait()
        return client

    def _new_mqtt_client(self) -> mqtt.Client:
        # Setup mqtt client
        client = mqtt.Client(client_id=self._client_id, protocol=mqtt.MQTTv311, clean_session=False)
        client.username_pw_set(username=self._cloud_creds.user_id, password=self._mqtt_password)

        # Certificate validation setup
        client.tls_set(
            ca_certs=self._ca_cert,
            certfile=None,
            keyfile=None,
            cert_reqs=ssl.CERT_NONE if self._mqtt_skip_validation else ssl.CERT_REQUIRED,
            tls_version=ssl.PROTOCOL_TLS_CLIENT,
            ciphers=None,
        )
        client.tls_insecure_set(self._mqtt_skip_validation)

        # Setup Callbacks
        client.on_connect = self._on_connect
        client.on_message = self._on_message
        client.on_disconnect = self._on_disconnect
        client.on_subscribe = self._on_subscribe

        return client

    def register_push_notification_handler_coroutine(
            self, coro: ManagerPushNotificationHandlerType
    ) -> None:
        """
        Registers a coroutine so that it gets invoked whenever a push notification is received from the Meross
        MQTT broker.
        :param coro: coroutine-function: a function that, when invoked, returns a Coroutine object that can be awaited.
        :return:
        """
        if not asyncio.iscoroutinefunction(coro):
            raise ValueError("The coro parameter must be a coroutine function")
        if coro in self._push_coros:
            _LOGGER.error(
                f"Coroutine {coro} was already added to event handlers of this device"
            )
            return
        self._push_coros.append(coro)

    def unregister_push_notification_handler_coroutine(
            self, coro: ManagerPushNotificationHandlerType
    ) -> None:
        """
        Unregisters the event handler
        :param coro: coroutine-function: a function that, when invoked, returns a Coroutine object that can be awaited.
                     This coroutine function should have been previously registered
        :return:
        """
        if coro in self._push_coros:
            self._push_coros.remove(coro)
        else:
            _LOGGER.error(
                f"Coroutine function {coro} was not registered as handler for this device"
            )

    def close(self):
        _LOGGER.info("Manager stop requested.")
        _LOGGER.debug("Canceling pending futures...")
        for f in _PENDING_FUTURES:
            if not f.cancelled():
                f.cancel()
        # Disconnect from all mqtt clients
        for client in self._mqtt_clients.values():
            client.disconnect()

    def find_devices(
            self,
            device_uuids: Optional[Iterable[str]] = None,
            internal_ids: Optional[Iterable[str]] = None,
            device_type: Optional[str] = None,
            device_class: Optional[Union[type, Iterable[type]]] = None,
            device_name: Optional[str] = None,
            online_status: Optional[OnlineStatus] = None,
    ) -> List[T]:
        """
        Lists devices that have been discovered via this manager. When invoked with no arguments,
        it returns the whole list of registered devices. When one or more filter arguments are specified,
        it returns the list of devices that satisfy all the filters (consider multiple filters as in logical AND).

        :param device_uuids: List of Meross native device UUIDs. When specified, only devices that have a native UUID
            contained in this list are returned.
        :param internal_ids: Iterable List of MerossIot device ids. When specified, only devices that have a
            derived-ids contained in this list are returned.
        :param device_type: Device type string as reported by meross app (e.g. "mss310" or "msl120"). Note that this
            field is case sensitive.
        :param device_class: Filter based on the resulting device class or list of classes. When this parameter is
            a list of types, the filter returns al the devices that matches at least one of the types in the list
            (logic OR). You can filter also for capability Mixins, such as
            :code:`meross_iot.controller.mixins.toggle.ToggleXMixin` (returns all the devices supporting ToggleX
            capability) or :code:`meross_iot.controller.mixins.light.LightMixin`
            (returns all the device that supports light control). Similarly, you can identify all the HUB devices
            by specifying :code:`meross_iot.controller.device.HubDevice`, Sensors as
            :code:`meross_iot.controller.subdevice.Ms100Sensor` and Valves as
            :code:`meross_iot.controller.subdevice.Mts100v3Valve`.
        :param device_name: Filter the devices based on their assigned name (case sensitive)
        :param online_status: Filter the devices based on their :code:`meross_iot.model.enums.OnlineStatus`
            as reported by the HTTP api or byt the relative hub (when dealing with subdevices).
        :return:
            The list of devices that match the provided filters, if any.
        """
        return self._device_registry.find_all_by(
            device_uuids=device_uuids,
            internal_ids=internal_ids,
            device_type=device_type,
            device_class=device_class,
            device_name=device_name,
            online_status=online_status,
        )

    async def async_device_discovery(
            self,
            update_subdevice_status: bool = True,
            meross_device_uuid: str = None,
            cached_http_device_list: Optional[Iterable[HttpDeviceInfo]] = None
    ) -> Iterable[BaseDevice]:
        """
        Fetch devices and online status from HTTP API. This method also notifies/updates local device online/offline
        status.

        :param meross_device_uuid: Meross UUID of the device that the user wants to discover (is already known).
            This parameter restricts the discovery only to that particular device.

        :param update_subdevice_status: When True, tells the manager to retrieve the HUB status in order to update
            hub-subdevice online status, which would be UNKNOWN if not explicitly retrieved.

        :param cached_http_device_list: List/Iterable structure of HttpDeviceInfo to be used for the discovery.
            When passed, the manger skips the HTTP API call and uses this data to perform MQTT discovery.
            When not passed, the manager will issue the HTTP API call to retrieve the latest HTTP devices list

        :return: A list of discovered device, which implement `BaseDevice`
        """
        if cached_http_device_list is None:
            _LOGGER.info(f"\n\n------- Triggering Manager Discovery, filter_device: [{meross_device_uuid}] -------")
            http_devices = await self._http_client.async_list_devices()
        else:
            _LOGGER.info(
                f"\n\n------- Triggering Manager Discovery (using cached http device list), filter_device: [{meross_device_uuid}] -------")
            http_devices = cached_http_device_list

        # If the user pased a specific uuid, filter the list by that one
        if meross_device_uuid is not None:
            http_devices = filter(lambda d: d.uuid == meross_device_uuid, http_devices)

        # Update state of local devices
        discovered_new_http_devices = []
        already_known_http_devices = {}
        for hdevice in http_devices:
            # Check if the device is already present into the registry
            ldevice = self._device_registry.lookup_base_by_uuid(hdevice.uuid)
            if ldevice is not None:
                already_known_http_devices[hdevice] = ldevice
            else:
                # If the http_device was not locally registered, keep track of it as we will add it later.
                discovered_new_http_devices.append(hdevice)

        # Give some info
        _LOGGER.debug(
            f"The following devices were already known to me: {already_known_http_devices}"
        )
        _LOGGER.debug(
            f"The following devices are new to me: {discovered_new_http_devices}"
        )

        enrolled_devices = []
        for d in discovered_new_http_devices:
            dev = await self._async_enroll_new_http_dev(d)
            enrolled_devices.append(dev)
        for hdevice, ldevice in already_known_http_devices.items():
            dev = await ldevice.update_from_http_state(hdevice)
            enrolled_devices.append(dev)

        _LOGGER.debug(
            f"Updating %d known devices form HTTPINFO and fetching "
            f"data from %d newly discovered devices...",
            len(already_known_http_devices),
            len(discovered_new_http_devices)
        )

        _LOGGER.info(f"Fetch and update done")

        hubs = []
        enrolled_subdevices = []
        for d in enrolled_devices:
            if isinstance(d, HubDevice):
                hubs.append(d)
                subdevs = await self._http_client.async_list_hub_subdevices(
                    hub_id=d.uuid
                )
                for sd in subdevs:
                    dev = await self._async_enroll_new_http_subdev(
                        subdevice_info=sd,
                        hub=d,
                        hub_reported_abilities=d.abilities)
                    enrolled_subdevices.append(dev)

        # We need to update the state of hubs in order to refresh subdevices online status
        if update_subdevice_status:
            for h in hubs:
                await h.async_update(drop_on_overquota=False)
        _LOGGER.info(f"\n------- Manager Discovery ended -------\n")

        res = []
        res.extend(enrolled_devices)
        res.extend(enrolled_subdevices)
        return res

    async def _async_enroll_new_http_subdev(
            self,
            subdevice_info: HttpSubdeviceInfo,
            hub: HubDevice,
            hub_reported_abilities: dict,
    ) -> Optional[GenericSubDevice]:
        subdevice = build_meross_subdevice(
            http_subdevice_info=subdevice_info,
            hub_uuid=hub.uuid,
            hub_reported_abilities=hub_reported_abilities,
            manager=self,
        )
        # Register the device to the hub
        if hub.get_subdevice(subdevice_id=subdevice.subdevice_id) is None:
            hub.register_subdevice(subdevice=subdevice)
        else:
            _LOGGER.debug("HUB %s already knows subdevice %s", hub.uuid, subdevice)

        # Enroll the device
        self._device_registry.enroll_device(subdevice)
        return subdevice

    async def async_init(self):
        """
        @deprecated
        Ignored, signature left for backward compatibility
        """
        pass

    async def _async_enroll_new_http_dev(
            self, device_info: HttpDeviceInfo
    ) -> Optional[BaseDevice]:
        # If the device is online, try to query the device for its abilities.
        device = None
        abilities = None
        if device_info.online_status == OnlineStatus.ONLINE:
            try:
                res_abilities = await self.async_execute_cmd(
                    destination_device_uuid=device_info.uuid,
                    method="GET",
                    namespace=Namespace.SYSTEM_ABILITY,
                    payload={},
                    mqtt_hostname=extract_domain(device_info.domain),
                    mqtt_port=DEFAULT_MQTT_PORT
                )
                abilities = res_abilities.get("ability")
            except CommandTimeoutError:
                _LOGGER.warning(
                    f"Device %s (%s) is online, but timeout occurred "
                    f"when fetching its abilities. ", str(device_info.dev_name), str(device_info.uuid)
                )
        if abilities is not None:
            # Build a full-featured device using the given ability set
            device = build_meross_device_from_abilities(
                http_device_info=device_info, device_abilities=abilities, manager=self
            )
        else:
            # In case we failed to build device's abilities at runtime, try to build the device statically
            # based on its model type.
            try:
                device = build_meross_device_from_known_types(
                    http_device_info=device_info, manager=self
                )
                _LOGGER.info(
                    "Device %s (%s) was built statically via known "
                    "types, because we failed to retrieve updated abilities for the given device.",
                    device_info.dev_name, str(device_info.uuid)
                )
            except UnknownDeviceType:
                _LOGGER.debug("Could not build statically device %s (%s) as it's not a known type.",
                              device_info.dev_name, device_info.uuid)

        # Enroll the device
        if device is not None:
            self._device_registry.enroll_device(device)
            return device

    def _on_connect(self, client: mqtt.Client, userdata, rc, other):
        # NOTE! This method is called by the paho-mqtt thread, thus any invocation to the
        # asyncio platform must be scheduled via `self._loop.call_soon_threadsafe()` method.
        topics = [(self._user_topic, 1), (self._client_response_topic, 1)]

        _LOGGER.debug(f"Connected with result code {rc}")
        # Subscribe to the relevant topics
        _LOGGER.debug("Subscribing to topics...")
        result, mid = client.subscribe(topics)

        if result != mqtt.MQTT_ERR_SUCCESS:
            _LOGGER.error("Failed to subscribe to topics %s", str(topics))

    def _on_disconnect(self, client: mqtt.Client, userdata, rc):
        # NOTE! This method is called by the paho-mqtt thread, thus any invocation to the
        # asyncio platform must be scheduled via `self._loop.call_soon_threadsafe()` method.
        _LOGGER.info("Disconnection detected. Reason: %s" % str(rc))

        # When a disconnection occurs, we need to set "unavailable" status.
        asyncio.run_coroutine_threadsafe(
            self._notify_connection_drop(), loop=self._loop
        )

        conn_evt = self._mqtt_connected_and_subscribed.get(userdata) # type: asyncio.Event
        conn_evt.clear()

    def _on_unsubscribe(self):
        # NOTE! This method is called by the paho-mqtt thread, thus any invocation to the
        # asyncio platform must be scheduled via `self._loop.call_soon_threadsafe()` method.
        _LOGGER.debug("Unsubscribed from topics")

    def _on_subscribe(self, client: mqtt.Client, userdata, mid, granted_qos):
        # NOTE! This method is called by the paho-mqtt thread, thus any invocation to the
        # asyncio platform must be scheduled via `self._loop.call_soon_threadsafe()` method.
        _LOGGER.debug("Successfully subscribed to topics.")
        sub_event = self._mqtt_connected_and_subscribed.get(userdata)
        self._loop.call_soon_threadsafe(sub_event.set)

        # When the connection happens after a disconnection (i.e. it is a re-connection)
        # we need to trigger Online Events for devices which where offline before.
        # Also, we want to update entirely the device status
        # To avoid flooding, schedule updates every 2s intervals.
        _LOGGER.info(
            "Subscribed to topics, scheduling state update for already known devices."
        )

        async def _update_devices_after_reconnection():
            # In case of reconnections, we need to issue a device_discovery via HTTP in order to update
            # ONLINE state. We also need to manually trigger "ONLINE" events for devices that resulted to be
            # OFFLINE and went ONLINE while our manager was off-network. To do so, we store into a dict the previous
            # online state and then issue the update only for devices that changed their state

            # Store the previous connection state for all known devices
            _prev_online_status = {d.uuid: d.online_status for d in self.find_devices()}

            # Issue a new discovery to update their connection status. This will rely on HTTP api to update it
            await self.async_device_discovery(update_subdevice_status=True)

            i = 0
            for d in self.find_devices():
                old_status = _prev_online_status.get(d.uuid)
                if old_status is None:
                    # This is a new device that has been added while we were offline.
                    _LOGGER.warning("Found a new device %s that has become online while we were offline.", d)
                    # TODO: do we need to issue a BINDING event manually here?
                    continue
                else:
                    _schedule_later(self._update_and_send_push(dev=d, old_status=old_status), start_delay=i,
                                    loop=self._loop)
                    i += 1

        # If a connection drop occurs, we must update the device state in order to be consistent
        # TODO: Do we need to issue this command only when connection drops occur or also at first connection attempt?
        if self._auto_discovery_on_connection:
            asyncio.run_coroutine_threadsafe(coro=_update_devices_after_reconnection(), loop=self._loop)

    async def _update_and_send_push(self, dev: BaseDevice, old_status: OnlineStatus) -> None:
        if dev.online_status == OnlineStatus.ONLINE:
            # In case the device was known and is ONLINE, we want to manually update its status
            _LOGGER.warning("Updating status for device %s", dev)
            await dev.async_update()

        # In case the device was known and is not ONLINE, we just send the ONLINE push notification
        if dev.online_status != old_status:
            _LOGGER.warning("Device %s changed its online status while manager was offline (was %s, now is %s). "
                            "Sending event manually.", dev, old_status, dev.online_status)
            await dev.async_handle_push_notification(namespace=Namespace.SYSTEM_ONLINE,
                                                     data={'online': {'status': dev.online_status.value}})

    def _on_message(self, client, userdata, msg):
        # NOTE! This method is called by the paho-mqtt thread, thus any invocation to the
        # asyncio platform must be scheduled via `self._loop.call_soon_threadsafe()` method.
        _LOGGER.debug(f"Received message from topic {msg.topic}: {str(msg.payload)}")

        # In order to correctly dispatch a message, we should look at:
        # - message destination topic
        # - message methods
        # - source device (from value in header)
        # Based on the network capture of Meross Devices, we know that there are 4 kinds of messages:
        # 1. COMMANDS sent from the app to the device (/appliance/<uuid>/subscribe) topic.
        #    Such commands have "from" header populated with "/app/<userid>-<appuuid>/subscribe" as that tells the
        #    device where to send its command ACK. Valid methods are GET/SET
        # 2. COMMAND-ACKS, which are sent back from the device to the app requesting the command execution on the
        #    "/app/<userid>-<appuuid>/subscribe" topic. Valid methods are GETACK/SETACK/ERROR
        # 3. PUSH notifications, which are sent to the "/app/46884/subscribe" topic from the device (which populates
        #    the from header with its topic /appliance/<uuid>/subscribe). In this case, only the PUSH
        #    method is allowed.
        # Case 1 is not of our interest, as we don't want to get notified when the device receives the command.
        # Instead we care about case 2 to acknowledge commands from devices and case 3, triggered when another app
        # has successfully changed the state of some device on the network.

        # Let's parse the message
        message = json.loads(str(msg.payload, "utf8"))
        header = message["header"]
        if not verify_message_signature(header, self._cloud_creds.key):
            _LOGGER.error(
                f"Invalid signature received. Message will be discarded. Message: {msg.payload}"
            )
            return

        _LOGGER.debug("Message signature OK")

        # Let's retrieve the destination topic, message method and source party:
        destination_topic = msg.topic
        message_method = header.get("method")
        source_topic = header.get("from")

        # Dispatch the message.
        # Check case 2: COMMAND_ACKS. In this case, we don't check the source topic address, as we trust it's
        # originated by a device on this network that we contacted previously.
        if destination_topic == build_client_response_topic(
                self._cloud_creds.user_id, self._app_id
        ) and message_method in ["SETACK", "GETACK", "ERROR"]:
            _LOGGER.debug("This message is an ACK to a command this client has send.")

            # If the message is a PUSHACK/GETACK/ERROR, check if there is any pending command waiting for it and, if so,
            # resolve its future
            message_id = header.get("messageId")
            future = self._pending_messages_futures.get(message_id)
            if future is not None:
                _LOGGER.debug("Found a pending command waiting for response message")
                if message_method == "ERROR":
                    err = CommandError(error_payload=message.get('payload'))
                    if not self._loop.is_closed():
                        self._loop.call_soon_threadsafe(_handle_future, future, None, err)
                    else:
                        _LOGGER.warning("Could not return message %s to caller as the event loop has been closed already", message)
                elif message_method in ("SETACK", "GETACK"):
                    if not self._loop.is_closed():
                        self._loop.call_soon_threadsafe(
                            _handle_future, future, message, None
                        )  # future.set_exception
                    else:
                        _LOGGER.warning("Could not return message %s to caller as the event loop has been closed already", message)
                else:
                    _LOGGER.error(
                        f"Unhandled message method {message_method}. Please report it to the developer."
                        f"raw_msg: {msg}"
                    )
                del self._pending_messages_futures[message_id]
        # Check case 3: PUSH notification.
        # Again, here we don't check the source topic, we trust that's legitimate.
        elif (
                destination_topic == build_client_user_topic(self._cloud_creds.user_id)
                and message_method == "PUSH"
        ):
            namespace = header.get("namespace")
            payload = message.get("payload")
            origin_device_uuid = device_uuid_from_push_notification(source_topic)

            parsed_push_notification = parse_push_notification(
                namespace=namespace,
                message_payload=payload,
                originating_device_uuid=origin_device_uuid,
            )
            if parsed_push_notification is None:
                _LOGGER.error(
                    "Push notification parsing failed. That message won't be dispatched."
                )
            else:
                asyncio.run_coroutine_threadsafe(
                    self._handle_and_dispatch_push_notification(
                        parsed_push_notification
                    ),
                    loop=self._loop,
                )
        else:
            _LOGGER.warning(
                f"The current implementation of this library does not handle messages received on topic "
                f"({destination_topic}) and when the message method is {message_method}. "
                "If you see this message many times, it means Meross has changed the way its protocol "
                "works. Contact the developer if that happens!"
            )

    async def _async_dispatch_push_notification(
            self, push_notification: GenericPushNotification
    ) -> bool:
        handled = False
        # Lookup the originating device and deliver the push notification to that one.
        # Exclude subdevices: event dispatching among them is handled by their relative HUB.
        target_devs = self._device_registry.find_all_by(
            device_uuids=(push_notification.originating_device_uuid,),
            exclude_classes=(GenericSubDevice,)
        )
        dev = None

        if len(target_devs) < 1:
            _LOGGER.warning(
                f"Received a push notification ({push_notification.namespace}, "
                f"raw_data: {json.dumps(push_notification.raw_data)}) for device(s) "
                f"({push_notification.originating_device_uuid}) that "
                f"are not available in the local registry. Trigger a discovery to intercept those events."
            )

        if len(target_devs) > 0:
            # Pass the control to the specific device implementation
            for dev in target_devs:
                try:
                    handled = (
                            await dev.async_handle_push_notification(
                                namespace=push_notification.namespace,
                                data=push_notification.raw_data,
                            )
                            or handled
                    )
                except Exception as e:
                    _LOGGER.exception(
                        "An unhandled exception occurred while handling push notification"
                    )

        else:
            _LOGGER.warning(
                "Received a push notification for a device that is not available in the local registry. "
                "You may need to trigger a discovery to catch those updates. Device-UUID: "
                f"{push_notification.originating_device_uuid}"
            )

        return handled

    async def _async_handle_push_notification_post_dispatching(
            self, push_notification: GenericPushNotification
    ) -> bool:
        if isinstance(push_notification, UnbindPushNotification):
            _LOGGER.info(
                "Received an Unbind PushNotification. Releasing device resources..."
            )
            devs = self._device_registry.find_all_by(
                device_uuids=(push_notification.originating_device_uuid)
            )
            for d in devs:
                _LOGGER.info(f"Releasing resources for device {d.internal_id}")
                self._device_registry.relinquish_device(
                    device_internal_id=d.internal_id
                )
            return True
        return False

    async def _handle_and_dispatch_push_notification(
            self, push_notification: GenericPushNotification
    ) -> None:
        """
        This method runs within the event loop and is responsible for handling and dispatching push notifications
        to the relative meross device within the registry.

        :param push_notification:
        :return:
        """
        # Dispatching
        handled_device = await self._async_dispatch_push_notification(
            push_notification=push_notification
        )

        # Notify any listener that registered explicitly to push_notification
        target_devs = self._device_registry.find_all_by(
            device_uuids=(push_notification.originating_device_uuid,)
        )

        for handler in self._push_coros:
            try:
                await handler(push_notification, target_devs, self)
            except Exception as e:
                _LOGGER.exception(f"Uncaught error occurred while executing push notification "
                                  f"handler {handler} for {push_notification}")

        # Handling post-dispatching
        handled_post = await self._async_handle_push_notification_post_dispatching(
            push_notification=push_notification
        )

        if not (handled_device or handled_post):
            _LOGGER.warning(
                f"Uncaught push notification {push_notification.namespace}. "
                f"Raw data: {json.dumps(push_notification.raw_data)}"
            )

    async def async_execute_cmd(
            self,
            mqtt_hostname: str,
            mqtt_port: int,
            destination_device_uuid: str,
            method: str,
            namespace: Union[Namespace, str],
            payload: dict,
            timeout: float = DEFAULT_COMMAND_TIMEOUT,
            override_transport_mode: TransportMode = None
    ):
        """
        This method sends a command to the device, locally via HTTP or via the MQTT Meross broker.

        :param mqtt_hostname: the mqtt broker hostname
        :param mqtt_port: the mqtt broker port
        :param destination_device_uuid:
        :param method: Can be GET/SET
        :param namespace: Command namespace
        :param payload: A dict containing the payload to be sent
        :param timeout: Maximum time interval in seconds to wait for the command-answer
        :param override_transport_mode: when set, overrides the manager transport mode
        :return:
        """

        # Only attempt local http communication if enabled via configuration.
        transport_mode = override_transport_mode if override_transport_mode is not None else self._default_transport_mode
        attempt_lan = transport_mode == TransportMode.LAN_HTTP_FIRST or transport_mode == TransportMode.LAN_HTTP_FIRST_ONLY_GET and method.upper() == 'GET'
        if attempt_lan:
            # Check if the LocalIP is available for the given device
            device = self._device_registry.lookup_base_by_uuid(destination_device_uuid)
            if device is None:
                _LOGGER.debug("Cannot issue command via LAN (http) against device with uuid %s as the device is not yet available on the registry", destination_device_uuid)
                attempt_lan = False
            elif device.lan_ip is None:
                _LOGGER.debug("Cannot issue command via LAN (http) against device with uuid %s as the device has not reported any internal LAN ip.", destination_device_uuid)
                attempt_lan = False
            elif self._error_budget_manager.is_out_of_budget(destination_device_uuid):
                _LOGGER.debug("Cannot issue command via LAN (http) against device with uuid %s as the device has no more error budget left.", destination_device_uuid)
                attempt_lan = False
            if attempt_lan:
                try:
                    # In case we succeed here, return the data we got.
                    # Otherwise, try again with MQTT.
                    _LOGGER.debug("Sending %s-%s command via HTTP to %s via %s", method, str(namespace), destination_device_uuid, device.lan_ip)
                    return await self._async_execute_cmd_http(device_ip=device.lan_ip,destination_device_uuid=destination_device_uuid,method=method,namespace=namespace,payload=payload,timeout=min(timeout, 1.0))
                except Exception as e:
                    _LOGGER.exception("An error occurred while attempting to send a message over internal LAN to device %s. Retrying with MQTT transport.", destination_device_uuid)
                    self._error_budget_manager.notify_error(destination_device_uuid)

        # Retrieve the mqtt client for the given domain:port broker
        if self._override_mqtt_server is not None:
            _LOGGER.debug("Overriding MQTT host/port as per manager parameter")
            mqtt_hostname = self._override_mqtt_server[0]
            mqtt_port = self._override_mqtt_server[1]

        _LOGGER.debug("Sending %s-%s command via MQTT to %s via %s:%d", method, str(namespace), destination_device_uuid,
                      mqtt_hostname, mqtt_port)
        client = await self._async_get_create_mqtt_client(domain=mqtt_hostname, port=mqtt_port)
        return await self.async_execute_cmd_client(client=client,
                                                   destination_device_uuid=destination_device_uuid,
                                                   method=method,
                                                   namespace=namespace,
                                                   payload=payload,
                                                   timeout=timeout)

    async def _async_execute_cmd_http(self,
                                      device_ip: str,
                                      destination_device_uuid: str,
                                      method: str,
                                      namespace: Union[Namespace,str],
                                      payload: dict,
                                      timeout: float = 10.0):
        # Send the message over the network
        # Build the mqtt message we will send to the broker
        message, message_id = self._build_mqtt_message(method, namespace, payload, destination_device_uuid)
        device: BaseDevice = self._device_registry.lookup_base_by_uuid(destination_device_uuid)

        async with ClientSession() as session:
            message_data = message_id
            decrypt_response = False
            if device.support_encryption():
                # Ensure we have correctly set the encryption key. If not, set it right away
                if not device.is_encryption_key_set():
                    device.set_encryption_key(uuid=device.uuid, mrskey=self._cloud_creds.key, mac=device.mac_address)
                # Encrypt the data
                message_data = device.encrypt(message)
                decrypt_response = True

            async with session.post(f"http://{device_ip}/config", data=message_data, timeout=timeout, headers=_DEFAULT_HEADERS) as response:
                response_data = await response.text("utf8")

            if decrypt_response:
                response_data = device.decrypt(response_data.encode("utf8")).decode("utf8")
                response_data = response_data.rstrip('\0')

            data = json.loads(response_data)
            return data.get("payload")

    async def async_execute_cmd_client(self,
                                       client: mqtt.Client,
                                       destination_device_uuid: str,
                                       method: str,
                                       namespace: Namespace,
                                       payload: dict,
                                       timeout: float = 10.0):
        # Send the message over the network
        # Build the mqtt message we will send to the broker
        message, message_id = self._build_mqtt_message(method, namespace, payload, destination_device_uuid)

        # Create a future and perform the send/waiting to a task
        fut = self._loop.create_future()
        self._pending_messages_futures[message_id] = fut
        response = await self._async_send_and_wait_ack(
            client=client,
            future=fut,
            target_device_uuid=destination_device_uuid,
            message=message,
            timeout=timeout
        )
        return response.get("payload")

    async def _async_send_and_wait_ack(
            self, client: mqtt.Client, future: Future, target_device_uuid: str, message: bytes, timeout: float,
    ):
        if not client.is_connected():
            raise Exception("MQTT client not connected.")

        client.publish(
            topic=build_device_request_topic(target_device_uuid), payload=message
        )
        try:
            return await asyncio.wait_for(future, timeout)
        except TimeoutError as e:
            domain, port = self._get_client_from_domain_port(client=client)
            _LOGGER.error(
                "Timeout occurred while waiting a response for message %s sent to device uuid "
                "%s. Timeout was: %f seconds. Mqtt Host: %s:%d.",
                str(message), str(target_device_uuid), timeout, domain, port)
            raise CommandTimeoutError(message=str(message), target_device_uuid=target_device_uuid, timeout=timeout)
        except CommandError as e:
            domain, port = self._get_client_from_domain_port(client=client)
            _LOGGER.error(
                "Error occurred while waiting a response for message %s sent to device uuid "
                "%s. Mqtt Host: %s:%d. Returned error: %s",
                str(message), str(target_device_uuid), domain, port, e.error_payload)
            raise

    async def _notify_connection_drop(self):
        for d in self._device_registry.find_all_by():
            pushn = OnlinePushNotification(originating_device_uuid=d.uuid, raw_data={'online': {'status': -1}})
            await self._handle_and_dispatch_push_notification(pushn)

    def _build_mqtt_message(self, method: str, namespace: Union[Namespace, str], payload: dict, destination_device_uuid: str):
        """
        Sends a message to the Meross MQTT broker, respecting the protocol payload.

        :param method:
        :param namespace:
        :param payload:
        :param destination_device_uuid:

        :return:
        """

        # Generate a random 16 byte string
        randomstring = "".join(
            random.SystemRandom().choice(string.ascii_uppercase + string.digits)
            for _ in range(16)
        )

        # Hash it as md5
        md5_hash = md5()
        md5_hash.update(randomstring.encode("utf8"))
        messageId = md5_hash.hexdigest().lower()
        timestamp = int(round(time()))

        # Hash the messageId, the key and the timestamp
        md5_hash = md5()
        strtohash = "%s%s%s" % (messageId, self._cloud_creds.key, timestamp)
        md5_hash.update(strtohash.encode("utf8"))
        signature = md5_hash.hexdigest().lower()

        if not isinstance(namespace, Namespace) and not isinstance(namespace, str):
            raise ValueError("Namespace parameter must be a Namespace enum or a string.")
        namespace_val = namespace.value if isinstance(namespace, Namespace) else namespace

        data = {
            "header": {
                "from": self._client_response_topic,
                "messageId": messageId,  # Example: "122e3e47835fefcd8aaf22d13ce21859"
                "method": method,  # Example: "GET",
                "namespace": namespace_val,  # Example: "Appliance.System.All",
                "payloadVersion": 1,
                "sign": signature,  # Example: "b4236ac6fb399e70c3d61e98fcb68b74",
                "timestamp": timestamp,
                "triggerSrc": "Android",
                "uuid": destination_device_uuid
            },
            "payload": payload,
        }

        strdata = json.dumps(data,separators=(',', ':'))
        return strdata.encode("utf-8"), messageId

    def set_proxy(self, proxy_type, proxy_addr, proxy_port):
        self._enable_proxy = True
        self._proxy_type = proxy_type
        self._proxy_addr = proxy_addr
        self._proxy_port = proxy_port

        for k, client in self._mqtt_clients.items():
            _LOGGER.info("Setting proxy configuration for client %s...", k)
            client.proxy_set(proxy_type=self._proxy_type, proxy_addr=self._proxy_addr, proxy_port=self._proxy_port)
            client.reconnect()

    def dump_device_registry(self, filename):
        """
        Save the current list of devices into a file so that you can later re-load it without issuing
        a discovery. **Note**: the stored information might become out-of-date or unvalidated. For instance,
        a device name might change over time, as its online status or any other info that is not immutable (as the UUID).
        Use this with caution!
        """
        self._device_registry.dump_to_file(filename)

    def load_devices_from_dump(self, filename):
        """Reload the registry info from a dump. **Note**: this will override all the currently discovered devices."""
        self._device_registry.load_from_dump(filename, manager=self)


class DeviceRegistry(object):
    def __init__(self):
        self._devices_by_internal_id = {}

    def clear(self) -> None:
        """Clear all the registered devices"""
        ids = [devid for devid in self._devices_by_internal_id]
        for devid in ids:
            self.relinquish_device(devid)

    def dump_to_file(self, filename: str)->None:
        """Dump the current device list to a file"""
        dumped_base_devices = [{'abilities': x.abilities, 'info': x.cached_http_info.to_dict()} for x in self._devices_by_internal_id.values() if not isinstance(x, GenericSubDevice)]
        with open(filename, "wt") as f:
            json.dump(dumped_base_devices, f, default=lambda x: x.isoformat() if isinstance(x, datetime) else x.value if(isinstance(x,OnlineStatus)) else 'Not-Serializable')

    def load_from_dump(self, filename: str, manager: MerossManager) -> None:
        """Load the device registry from a file"""
        dumped_json_data = []
        with open(filename, "rt") as f:
            dumped_json_data = json.load(f)

        for deviced in dumped_json_data:
            device_abilities = deviced['abilities']
            device_info = HttpDeviceInfo.from_dict(deviced['info'])
            device = build_meross_device_from_abilities(http_device_info=device_info, device_abilities=device_abilities, manager=manager)
            self.enroll_device(device)

    def relinquish_device(self, device_internal_id: str):
        dev = self._devices_by_internal_id.get(device_internal_id)
        if dev is None:
            raise ValueError(
                f"Cannot relinquish device {device_internal_id} as it does not belong to this registry."
            )

        # Dismiss the device
        _LOGGER.debug(f"Disposing resources for {dev.name} ({dev.uuid})")
        dev.dismiss()
        del self._devices_by_internal_id[device_internal_id]
        _LOGGER.info(f"Device {dev.name} ({dev.uuid}) removed from registry")

    def enroll_device(self, device: BaseDevice):
        if device.internal_id in self._devices_by_internal_id:
            _LOGGER.info(
                f"Device {device.name} ({device.internal_id}) has been already added to the registry."
            )
            return
        else:
            _LOGGER.debug(
                f"Adding device {device.name} ({device.internal_id}) to registry."
            )
            self._devices_by_internal_id[device.internal_id] = device

    def lookup_by_id(self, device_id: str) -> Optional[BaseDevice]:
        return self._devices_by_internal_id.get(device_id)

    def lookup_base_by_uuid(self, device_uuid: str) -> Optional[BaseDevice]:
        res = list(
            filter(
                lambda d: d.uuid == device_uuid and not isinstance(d, GenericSubDevice),
                self._devices_by_internal_id.values(),
            )
        )
        if len(res) > 1:
            raise ValueError(f"Multiple devices found for device_uuid {device_uuid}")
        elif len(res) == 1:
            return res[0]
        else:
            return None

    def find_all_by(
            self,
            device_uuids: Optional[Iterable[str]] = None,
            internal_ids: Optional[Iterable[str]] = None,
            device_type: Optional[str] = None,
            device_class: Optional[Union[type, Iterable[type]]] = None,
            device_name: Optional[str] = None,
            online_status: Optional[OnlineStatus] = None,
            exclude_classes: Optional[Iterable[type]] = None
    ) -> List[BaseDevice]:

        def filter_by_type(dev: Any):
            for t in device_class:
                if isinstance(dev, t):
                    return True
            return False

        def filter_by_excluded_type(dev: Any):
            for t in exclude_classes:
                if isinstance(dev, t):
                    return False
            return True

        # Look by Internal UUIDs
        if internal_ids is not None:
            res = filter(
                lambda d: d.internal_id in internal_ids,
                self._devices_by_internal_id.values(),
            )
        else:
            res = self._devices_by_internal_id.values()

        if device_uuids is not None:
            res = filter(lambda d: d.uuid in device_uuids, res)
        if device_type is not None:
            res = filter(lambda d: d.type == device_type, res)
        if online_status is not None:
            res = filter(lambda d: d.online_status == online_status, res)
        if device_class is not None:
            if isinstance(device_class, type):
                res = filter(lambda d: isinstance(d, device_class), res)
            elif isinstance(device_class, Iterable):
                res = filter(filter_by_type, res)
        if device_name is not None:
            res = filter(lambda d: d.name == device_name, res)
        if exclude_classes is not None:
            res = filter(filter_by_excluded_type, res)

        return list(res)


def _handle_future(future: Future, result: object, exception: Exception):
    if future.cancelled():
        return

    if exception is not None:
        future.set_exception(exception)
    else:
        if future.cancelled():
            _LOGGER.debug("Skipping set_result for cancelled future.")
        elif future.done():
            _LOGGER.error("This future is already done: cannot set result.")
        else:
            future.set_result(result)


def set_future_done(coroutine, future):
    coroutine.close()
    if future in _PENDING_FUTURES:
        _PENDING_FUTURES.remove(future)


async def delayed_execution(coro, delay):
    await asyncio.sleep(delay=delay)
    await coro


def _schedule_later(coroutine, start_delay, loop):
    future = asyncio.run_coroutine_threadsafe(coro=delayed_execution(coro=coroutine, delay=start_delay), loop=loop)
    _PENDING_FUTURES.append(future)
    future.add_done_callback(functools.partial(set_future_done,coroutine))



import logging
from typing import Optional, Union

from meross_iot.model.enums import Namespace, get_or_parse_namespace
from meross_iot.model.push.bind import BindPushNotification
from meross_iot.model.push.generic import GenericPushNotification
from meross_iot.model.push.online import OnlinePushNotification
from meross_iot.model.push.unbind import UnbindPushNotification

_LOGGER = logging.getLogger(__name__)

_PUSH_NOTIFICATION_BINDING = {
    Namespace.CONTROL_BIND: BindPushNotification,
    Namespace.CONTROL_UNBIND: UnbindPushNotification,
    Namespace.SYSTEM_ONLINE: OnlinePushNotification
}


def parse_push_notification(namespace: Union[str, Namespace],
                            message_payload: dict,
                            originating_device_uuid: str,
                            ) -> Optional[GenericPushNotification]:
    """
    Parse the push notification and return the most appropriate handling class.
    The current implementation is able to discriminate Bind/Unbind specific push notification. All the
    others are returned as a GenericPushNotification.
    :param namespace:
    :param message_payload:
    :param originating_device_uuid:
    :return:
    """
    _LOGGER.debug(f"Parsing push notification {namespace}, payload: {message_payload}")

    # Parse the namespace
    try:
        parsed_namespace = get_or_parse_namespace(namespace)
        push_notification_class = _PUSH_NOTIFICATION_BINDING.get(parsed_namespace)  # type: type
        if push_notification_class is None:
            return GenericPushNotification(namespace=parsed_namespace, originating_device_uuid=originating_device_uuid,
                                           raw_data=message_payload)
        else:
            return push_notification_class(raw_data=message_payload,
                                           originating_device_uuid=originating_device_uuid)

    except ValueError:
        return None



import json
import logging
from datetime import datetime
from typing import Union, List

from meross_iot.model.constants import DEFAULT_MQTT_PORT, DEFAULT_MQTT_HOST
from meross_iot.model.enums import OnlineStatus
from meross_iot.model.shared import BaseDictPayload
from meross_iot.utilities.network import extract_domain, extract_port

_LOGGER = logging.getLogger(__name__)


class HttpDeviceInfo(BaseDictPayload):
    def __init__(self,
                 uuid: str,
                 online_status: Union[int, OnlineStatus],
                 dev_name: str,
                 device_type: str,
                 channels: List[dict],
                 fmware_version: str,
                 hdware_version: str,
                 domain: str,
                 reserved_domain: str,
                 sub_type: str = None,
                 bind_time: Union[int, datetime] = None,
                 skill_number: str = None,
                 user_dev_icon: str = None,
                 icon_type: int = None,
                 region: str = None,
                 dev_icon_id: str = None,
                 *args, **kwargs):

        super().__init__(*args, **kwargs)
        self.uuid = uuid
        if isinstance(online_status, int):
            self.online_status = OnlineStatus(online_status)
        elif isinstance(online_status, OnlineStatus):
            self.online_status = online_status
        else:
            _LOGGER.warning(f"Provided online_status is not int neither OnlineStatus. It will be ignored.")
            self.online_status = None

        self.dev_name = dev_name
        self.dev_icon_id = dev_icon_id
        if isinstance(bind_time, int):
            self.bind_time = datetime.utcfromtimestamp(bind_time)
        elif isinstance(bind_time, datetime):
            self.bind_time = bind_time
        elif isinstance(bind_time, str):
            self.bind_time = datetime.strptime(bind_time, "%Y-%m-%dT%H:%M:%S")
        else:
            _LOGGER.warning(f"Provided bind_time is not int neither datetime. It will be ignored.")
            self.bind_time = None

        self.device_type = device_type
        self.sub_type = sub_type
        self.channels = channels
        self.region = region
        self.fmware_version = fmware_version
        self.hdware_version = hdware_version
        self.user_dev_icon = user_dev_icon
        self.icon_type = icon_type
        self.skill_number = skill_number
        self.domain = domain
        self.reserved_domain = reserved_domain

    def get_mqtt_host(self) -> str:
        """Infers the mqtt server host for this device"""
        # Prefer domain over reserved domain
        if self.domain is not None:
            return extract_domain(self.domain)
        if self.reserved_domain is not None:
            return extract_domain(self.reserved_domain)
        _LOGGER.warning("Could not find domain info for device %s, returning default domain %s", str(self.uuid), DEFAULT_MQTT_HOST)
        return DEFAULT_MQTT_HOST

    def get_mqtt_port(self) -> int:
        """Infers the mqtt server port for this device"""
        # Prefer domain over reserved domain
        if self.domain is not None:
            return extract_port(self.domain, DEFAULT_MQTT_PORT)
        if self.reserved_domain is not None:
            return extract_port(self.reserved_domain, DEFAULT_MQTT_PORT)
        _LOGGER.warning("Could not find domain info for device %s, returning default port %d", str(self.uuid),
                    DEFAULT_MQTT_PORT)
        return DEFAULT_MQTT_PORT

    def __repr__(self):
        return json.dumps(self.__dict__, default=lambda x: x.isoformat() if isinstance(x,datetime) else x.name if(isinstance(x,OnlineStatus)) else "NOT-SERIALIZABLE")

    def __str__(self):
        basic_info = f"{self.dev_name} ({self.device_type}, HW {self.hdware_version}, FW {self.fmware_version})"
        return basic_info


from enum import Enum


class ErrorCodes(Enum):
    """
    Status codes returned by the Meross HTTP APIs
    """

    CODE_NO_ERROR = 0
    """Not an error"""

    CODE_MISSING_USER = 1000
    """Wrong or missing user"""

    CODE_MISSING_PASSWORD = 1001
    """Wrong or missing password"""

    CODE_UNEXISTING_ACCOUNT = 1002
    """Account does not exist"""

    CODE_DISABLED_OR_DELETED_ACCOUNT = 1003
    """This account has been disabled or deleted"""

    CODE_WRONG_CREDENTIALS = 1004
    """Wrong email or password"""

    CODE_INVALID_EMAIL = 1005
    """Invalid email address"""

    CODE_BAD_PASSWORD_FORMAT = 1006
    """Bad password format"""

    USER_ALREADY_EXISTS = 1007
    """User already exists"""

    CODE_WRONG_EMAIL = 1008
    """This email is not registered"""

    SEND_EMAIL_FAILED = 1009
    """Email send failed"""

    WRONG_TICKET = 1011
    """Wrong Ticket"""

    CODE_OVERDUE = 1012
    """Code Overdue"""

    WRONG_CODE = 1013
    """Wrong Code"""

    DUPLICATE_PASSWORD = 1014
    """Duplicate password"""

    SAME_EMAIL = 1015
    """Same email when changing account email"""

    CODE_TOKEN_INVALID = 1019
    """Token expired"""

    CODE_UNKNOWN_FAILURE_1021 = 1021
    """Unknown error"""

    CODE_TOKEN_ERROR = 1022
    """Token error"""

    CODE_UNKNOWN_FAILURE_1023 = 1023
    """Unknown error"""

    CODE_UNKNOWN_FAILURE_1024 = 1024
    """Unknown error"""

    CODE_UNKNOWN_FAILURE_1025 = 1025
    """Unknown error"""

    CODE_UNKNOWN_FAILURE_1026 = 1026
    """Unknown error"""

    CODE_UNKNOWN_FAILURE_1027 = 1027
    """Unknown error"""

    REQUESTED_TOO_FREQUENTLY = 1028
    """Requested too frequently"""

    CODE_REDIRECT_REGION = 1030
    """Wrong login region"""

    USER_NAME_NOT_MATCHING = 1031
    """Username does not match"""

    WRONG_MFA_CODE = 1032
    """Wrong MFA Code"""

    MFA_CODE_REQUIRED = 1033
    """MFA Code required"""

    OPERATION_IS_LOCKED = 1035
    """Operation is locked"""

    REPEAT_CHECK_IN = 1041
    """Repeat checkin"""

    API_TOP_LIMIT_REACHED = 1042
    """API Top limit reached"""

    RESOURCE_ACCESS_DENY = 1043
    """Resource access deny"""

    CODE_TOKEN_EXPIRED = 1200
    """Token has expired"""

    SERVER_UNABLE_GEN_TOKEN = 1201
    """Server was unable to generate token"""

    CODE_UNKNOWN_FAILURE_1202 = 1202
    """Unknown error"""

    CODE_UNKNOWN_FAILURE_1203 = 1203
    """Unknown error"""

    CODE_UNKNOWN_FAILURE_1204 = 1204
    """Unknown error"""

    CODE_UNKNOWN_FAILURE_1210 = 1210
    """Unknown error"""

    CODE_UNKNOWN_FAILURE_1211 = 1211
    """Unknown error"""

    CODE_UNKNOWN_FAILURE_1212 = 1212
    """Unknown error"""

    CODE_UNKNOWN_FAILURE_1213 = 1213
    """Unknown error"""

    CODE_UNKNOWN_FAILURE_1214 = 1214
    """Unknown error"""

    CODE_UNKNOWN_FAILURE_1215 = 1215
    """Unknown error"""

    CODE_UNKNOWN_FAILURE_1226 = 1226
    """Unknown error"""

    CODE_UNKNOWN_FAILURE_1227 = 1227
    """Unknown error"""

    CODE_UNKNOWN_FAILURE_1228 = 1228
    """Unknown error"""

    CODE_UNKNOWN_FAILURE_1229 = 1229
    """Unknown error"""

    CODE_UNKNOWN_FAILURE_1230 = 1230
    """Unknown error"""

    CODE_UNKNOWN_FAILURE_1231 = 1231
    """Unknown error"""

    CODE_UNKNOWN_FAILURE_1232 = 1232
    """Unknown error"""

    CODE_UNKNOWN_FAILURE_1233 = 1233
    """Unknown error"""

    CODE_MAX_CONTROL_BOARDS_REACHED = 1255
    """The number of remote control boards exceeded the limit"""

    CODE_COMPATIBILE_MODE_HAVING = 1256
    """Compatible mode having"""

    CODE_COMPATIBILE_MODE_NOT_HAVING = 1257
    """Compatible mode not having"""

    CODE_TOO_MANY_TOKENS = 1301
    """Too many tokens have been issued"""

    CODE_UNKNOWN_FAILURE_1400 = 1400
    """Unknown error"""

    CODE_UNKNOWN_FAILURE_1401 = 1401
    """Unknown error"""

    CODE_UNKNOWN_FAILURE_1402 = 1402
    """Unknown error"""

    CODE_UNKNOWN_FAILURE_1403 = 1403
    """Unknown error"""

    CODE_UNKNOWN_FAILURE_1500 = 1500
    """Unknown error"""

    CODE_UNKNOWN_FAILURE_1501 = 1501
    """Unknown error"""

    CODE_UNKNOWN_FAILURE_1502 = 1502
    """Unknown error"""

    CODE_UNKNOWN_FAILURE_1503 = 1503
    """Unknown error"""

    CODE_UNKNOWN_FAILURE_1504 = 1504
    """Unknown error"""

    CODE_UNKNOWN_FAILURE_1601 = 1601
    """Unknown error"""

    CODE_UNKNOWN_FAILURE_1602 = 1602
    """Unknown error"""

    CODE_UNKNOWN_FAILURE_1603 = 1603
    """Unknown error"""

    CODE_UNKNOWN_FAILURE_1604 = 1604
    """Unknown error"""

    CODE_UNKNOWN_FAILURE_1605 = 1605
    """Unknown error"""

    CODE_UNKNOWN_FAILURE_1700 = 1700
    """Unknown error"""

    CODE_UNSUPPORTED_TIMEZONE = 500
    """The selected timezone is not supported"""

    CODE_GENERIC_ERROR = 5000
    """Unknown or generic error"""

    CODE_UNKNOWN_FAILURE_5001 = 5001
    """Unknown or generic error"""

    CODE_UNKNOWN_FAILURE_5002 = 5002
    """Unknown or generic error"""

    CODE_UNKNOWN_FAILURE_5003 = 5003
    """Unknown or generic error"""

    CODE_UNKNOWN_FAILURE_5004 = 5004
    """Unknown or generic error"""

    CODE_IR_DEVICE_BUSY = 5020
    """Infrared Remote device is busy"""

    CODE_IR_RECORD_TIMEOUT = 5021
    """Infrared record timeout"""

    CODE_IR_RECORD_INVALID = 5022
    """Infrared record invalid"""

    SYSTEM_ERROR = 10001
    """System error"""

    UNKNOWN_ERROR = 10002
    """Unknown error"""

    SERIALIZE_ERROR = 10003
    """Serialize error"""

    HTTP_COMMON_ERROR = 10006
    """Http common error"""

    INVALID_PARAMETER = 20101
    """Invalid parameter"""

    RESOURCE_DOES_NOT_EXIST = 20106
    """Not existing resource"""

    UNSUPPORTED = 20112
    """Unsupported"""

    SEND_EMAIL_LIMIT = 20115
    """Send email limit"""



import asyncio
import json
import logging
import ssl
from hashlib import md5
from typing import Any, Dict, Tuple

from paho.mqtt.client import Client, MQTTv311, MQTTMessage

from meross_iot.utilities.mqtt import build_device_request_topic
from utilities.mixedqueue import MixedQueue

_LOGGER = logging.getLogger()


class FakeDeviceSniffer:
    _mqtt_client: Client

    def __init__(self, uuid: str, mac_address: str, meross_user_id: str, meross_cloud_key: str, mqtt_host: str, mqtt_port: int, logger: logging.Logger):
        """Constructor"""
        self._loop = asyncio.get_running_loop()
        self._l = logger
        self._uuid = uuid.lower()
        self._mac_address = mac_address
        self._meross_user_id = meross_user_id
        self._meross_cloud_key = meross_cloud_key
        self._mqtt_host = mqtt_host
        self._mqtt_port = mqtt_port
        self._msg_queue = MixedQueue(self._loop)

        # Build the client-id
        self._client_id = f"fmware:{self._uuid}_random"

        # Start the mqtt client and connect
        self._mqtt_client = Client(client_id=self._client_id, clean_session=True, userdata=None,
                                          protocol=MQTTv311, transport="tcp", reconnect_on_failure=False)
        mac_key_digest = md5(f"{self._mac_address}{self._meross_cloud_key}".encode("utf8")).hexdigest().lower()
        device_password = f"{self._meross_user_id}_{mac_key_digest}"
        self._mqtt_client.username_pw_set(username=self._mac_address, password=device_password)

        # Set up the handlers
        self._mqtt_client.on_connect = self._on_connect
        self._mqtt_client.on_disconnect = self._on_disconnect
        self._mqtt_client.on_subscribe = self._on_subscribe
        self._mqtt_client.on_connect_fail = self._on_connection_fail
        self._mqtt_client.on_unsubscribe = self._on_subscribe
        self._mqtt_client.on_message = self._on_message

        # Prepare synchronization events
        self._started = False
        self._starting = False
        self._connected_event = asyncio.Event()
        self._subscribed_event = asyncio.Event()
        self._disconnected_event = asyncio.Event()

        # Device topic
        self._device_topic = build_device_request_topic(client_uuid=self._uuid)

    async def async_start(self, timeout: float) -> None:
        """Starts the emulation"""
        if self._started or self._starting:
            self._l.error("The fake device has been already started")
            raise RuntimeError("Already Started.")

        self._starting = True
        self._mqtt_client.tls_set(ca_certs=None, certfile=None,
                                  keyfile=None, cert_reqs=ssl.CERT_REQUIRED,
                                  tls_version=ssl.PROTOCOL_TLS_CLIENT,
                                  ciphers=None)
        self._mqtt_client.connect(host=self._mqtt_host, port=self._mqtt_port)
        self._mqtt_client.loop_start()
        await asyncio.wait_for(self._connected_event.wait(), timeout)
        self._mqtt_client.subscribe(topic=self._device_topic)

    async def async_stop(self):
        if not self._started and not self._starting:
            raise RuntimeError("Not running.")
        if self._mqtt_client.is_connected():
            self._mqtt_client.disconnect()
            await self._disconnected_event.wait()
        self._mqtt_client.loop_stop()
        self._started = False

    def _on_connect(self, client: Client, userdata: Any, flags, rc):
        self._l.debug("Fake device connected to mqtt broker successfully")
        self._loop.call_soon_threadsafe(self._connected_event.set)

    def _on_subscribe(self, client: Client, userdata: Any, mid, granted_qos):
        self._l.debug("Fake device subscribed to mqtt topics successfully")
        def _cb():
            self._subscribed_event.set()
            self._starting = False
            self._started = True
        self._loop.call_soon_threadsafe(_cb)

    def _on_connection_fail(self, client: Client, userdata):
        def _cb():
            self._connected_event.clear()
            self._starting = False
            self._started = False
        self._l.error("Fake device failed to connect to MQTT broker")
        self._loop.call_soon_threadsafe(_cb)

    def _on_disconnect(self, client: Client, userdata, rc):
        def _cb():
            self._connected_event.clear()
            self._disconnected_event.set()
        self._l.info("Fake device disconnected from MQTT broker")
        self._loop.call_soon_threadsafe(_cb)

    def _on_unsubscribe(self, client: Client, userdata, mid):
        self._l.info("Fake device unsubscribed from MQTT topics")
        self._loop.call_soon_threadsafe(self._subscribed_event.clear)

    def _on_message(self, client: Client, userdata, message: MQTTMessage):
        self._l.info("Fake device received message: %s from topic %s", str(message.payload), str(message.topic))
        _LOGGER.info(msg=str(message.payload))
        self._msg_queue.sync_put(message)

    async def async_wait_for_message(self, valid_methods=('SET', 'GET')) -> Tuple[MQTTMessage, str, str, Dict]:
        while True:
            raw_message: MQTTMessage = await self._msg_queue.async_get()
            parsed_message = json.loads(str(raw_message.payload, "utf8"))
            namespace = parsed_message['header']['namespace']
            method = parsed_message['header']['method']
            payload = parsed_message['payload']

            # Discard ACKs and PUSH notifications
            if method in valid_methods:
                return raw_message, namespace, method, payload



import logging
from typing import Optional, List, Dict

from meross_iot.controller.device import ChannelInfo
from meross_iot.model.enums import Namespace, ThermostatMode

_LOGGER = logging.getLogger(__name__)


_THERMOSTAT_MIN_SETTABLE_TEMP = 5.0
_THERMOSTAT_MAX_SETTABLE_TEMP = 35


class ThermostatState:
    _state: Dict
    def __init__(self, state: Dict = None):
        if state is not None:
            self._state = state

    def update(self, state: Dict):
        """
        Updates the internal state from the raw data dictionary passed as parameter
        :param state:
        :return:
        """
        self._state.update(state)

    @property
    def is_on(self) -> Optional[bool]:
        """Whether the device is on or not."""
        on = self._state.get('onoff')
        if on is None:
            return None
        return on == 1

    @property
    def mode(self) -> Optional[ThermostatMode]:
        """The current thermostat mode"""
        mode = self._state.get('mode')
        if mode is None:
            return None
        return ThermostatMode(mode)

    @property
    def warning(self) -> Optional[bool]:
        """The warning state of the thermostat"""
        warn = self._state.get('warning')
        if warn is None:
            return None
        return warn == 1

    @property
    def target_temperature_celsius(self) -> Optional[float]:
        """The target temperature of the thermostat"""
        temp = self._state.get('targetTemp')
        if temp is None:
            return None
        return float(temp)/10.0

    @property
    def min_temperature_celsius(self) -> Optional[float]:
        """The minimum settable temperature for the thermostat"""
        temp = self._state.get('min')
        if temp is None:
            return None
        return float(temp)/10.0

    @property
    def max_temperature_celsius(self) -> Optional[float]:
        """The maximum settable temperature for the thermostat"""
        temp = self._state.get('max')
        if temp is None:
            return None
        return float(temp)/10.0

    @property
    def current_temperature_celsius(self) -> Optional[float]:
        """The current ambient temperature"""
        temp = self._state.get('currentTemp')
        if temp is None:
            return None
        return float(temp)/10.0

    @property
    def heat_temperature_celsius(self) -> Optional[float]:
        """The target temperature when HEAT mode is enabled"""
        temp = self._state.get('heatTemp')
        if temp is None:
            return None
        return float(temp)/10.0

    @property
    def cool_temperature_celsius(self) -> Optional[float]:
        """The target temperature when COOL mode is enabled"""
        temp = self._state.get('coolTemp')
        if temp is None:
            return None
        return float(temp)/10.0

    @property
    def eco_temperature_celsius(self) -> Optional[float]:
        """The target temperature when ECO mode is enabled"""
        temp = self._state.get('ecoTemp')
        if temp is None:
            return None
        return float(temp)/10.0

    @property
    def manual_temperature_celsius(self) -> Optional[float]:
        """The target temperature when AUTO mode is enabled"""
        temp = self._state.get('manualTemp')
        if temp is None:
            return None
        return float(temp)/10.0


class ThermostatModeMixin:
    _execute_command: callable
    check_full_update_done: callable
    _thermostat_state_by_channel: Dict[int, ThermostatState]

    def __init__(self, device_uuid: str,
                 manager,
                 **kwargs):
        super().__init__(device_uuid=device_uuid, manager=manager, **kwargs)
        self._thermostat_state_by_channel = {}

    def _update_mode(self, mode_data: Dict):
        # The MTS200 thermostat does bring a object for every sensor/channel it handles.
        for c in mode_data:
            channel_index = c['channel']
            state = self._thermostat_state_by_channel.get(channel_index)
            if state is None:
                state = ThermostatState(c)
                self._thermostat_state_by_channel[channel_index] = state
            else:
                state.update(c)

    async def async_handle_push_notification(self, namespace: Namespace, data: dict) -> bool:
        locally_handled = False

        if namespace == Namespace.CONTROL_THERMOSTAT_MODE:
            _LOGGER.debug(f"{self.__class__.__name__} handling push notification for namespace "
                          f"{namespace}")
            mode_data = data.get('mode')
            if mode_data is None:
                _LOGGER.error(f"{self.__class__.__name__} could not find 'mode' attribute in push notification data: "
                              f"{data}")
                locally_handled = False
            else:
                self._update_mode(mode_data)
                locally_handled = True

        # Always call the parent handler when done with local specific logic. This gives the opportunity to all
        # ancestors to catch all events.
        parent_handled = await super().async_handle_push_notification(namespace=namespace, data=data)
        return locally_handled or parent_handled

    async def async_handle_update(self, namespace: Namespace, data: dict) -> bool:
        _LOGGER.debug(f"Handling {self.__class__.__name__} mixin data update.")
        locally_handled = False
        if namespace == Namespace.SYSTEM_ALL:
            thermostat_data = data.get('all', {}).get('digest', {}).get('thermostat', {})
            mode_data = thermostat_data.get('mode')
            if mode_data is not None:
                self._update_mode(mode_data)
            locally_handled = True

        super_handled = await super().async_handle_update(namespace=namespace, data=data)
        return super_handled or locally_handled

    def get_thermostat_state(self, channel: int = 0, *args, **kwargs) -> Optional[ThermostatState]:
        """
        Returns the current thermostat state
        :param channel:
        :param args:
        :param kwargs:
        :return:
        """
        self.check_full_update_done()
        state = self._thermostat_state_by_channel.get(channel)
        return state

    def _align_temp(self, temperature:float, channel: int = 0) -> float:
        """
            Given an input temperature for a specific channel, checks if the temperature is within the ranges
            of acceptable values and rounds it to the nearest 0.5 value. It also applies the 10x multiplication
            as Meross devices requires decimal-degrees
        """
        # Retrieve the min/max settable values from the state.
        # If this is not available, assume some defaults
        channel_state = self._thermostat_state_by_channel.get(channel)
        min_settable_temp = _THERMOSTAT_MIN_SETTABLE_TEMP
        max_settable_temp = _THERMOSTAT_MIN_SETTABLE_TEMP
        if channel_state is not None:
            min_settable_temp = channel_state.min_temperature_celsius
            max_settable_temp = channel_state.max_temperature_celsius

        if temperature < min_settable_temp or temperature > max_settable_temp:
            raise ValueError("The provided temperature value is invalid or out of range for this device.")

        # Round temp value to 0.5
        quotient = temperature/0.5
        quotient = round(quotient)
        final_temp = quotient*5
        return final_temp

    async def async_set_thermostat_config(self,
                                    channel: int = 0,
                                    mode: Optional[ThermostatMode] = None,
                                    manual_temperature_celsius: Optional[float] = None,
                                    heat_temperature_celsius: Optional[float] = None,
                                    cool_temperature_celsius: Optional[float] = None,
                                    eco_temperature_celsius: Optional[float] = None,
                                    on_not_off: Optional[bool] = None,
                                    timeout: Optional[float] = None,
                                    *args,
                                    **kwargs) -> None:
        channel_conf = {
            'channel': channel
        }
        payload = {'mode': [channel_conf]}

        # Arg check
        if mode is not None:
            channel_conf['mode'] = mode.value
        if manual_temperature_celsius is not None:
            channel_conf['manualTemp'] = self._align_temp(manual_temperature_celsius, channel=channel)
        if heat_temperature_celsius is not None:
            channel_conf['heatTemp'] = self._align_temp(heat_temperature_celsius, channel=channel)
        if cool_temperature_celsius is not None:
            channel_conf['coolTemp'] = self._align_temp(cool_temperature_celsius, channel=channel)
        if eco_temperature_celsius is not None:
            channel_conf['ecoTemp'] = self._align_temp(eco_temperature_celsius, channel=channel)
        if on_not_off is not None:
            channel_conf['onoff'] = 1 if on_not_off else 0

        # This api call will return the updated state of the device. We use it to update the internal state right away.
        result = await self._execute_command(method="SET",
                              namespace=Namespace.CONTROL_THERMOSTAT_MODE,
                                         payload=payload,
                                         timeout=timeout)
        mode_data = result.get('mode')
        self._update_mode(mode_data)



import os
from typing import List

from aiohttp import web
from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop

from meross_iot.controller.device import BaseDevice
from meross_iot.controller.mixins.garage import GarageOpenerMixin
from meross_iot.controller.mixins.toggle import ToggleXMixin
from meross_iot.manager import MerossManager
from meross_iot.model.enums import OnlineStatus
from meross_iot.model.push.generic import GenericPushNotification
from tests import async_get_client


if os.name == 'nt':
    import asyncio
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
else:
    import asyncio


class TestPushNotificationHandler(AioHTTPTestCase):
    async def get_application(self):
        return web.Application()

    async def setUpAsync(self):
        # Wait some time before next test-burst
        await asyncio.sleep(10)
        self.meross_client, self.requires_logout = await async_get_client()

        # Look for a device to be used for this test
        self.meross_manager = MerossManager(http_client=self.meross_client)
        await self.meross_manager.async_init()
        devices = await self.meross_manager.async_device_discovery()
        toggle_devices = self.meross_manager.find_devices(device_class=ToggleXMixin, online_status=OnlineStatus.ONLINE)
        # Exclude garage openers
        toggle_devices = list(filter(lambda x: not isinstance(x, GarageOpenerMixin), toggle_devices))
        if len(toggle_devices) < 1:
            self.test_device = None
        else:
            self.test_device = toggle_devices[0]

    @unittest_run_loop
    async def test_dev_push_notification(self):
        if self.test_device is None:
            self.skipTest("No ToggleX device has been found to run this test on.")
        print(f"Testing device {self.test_device.name}")
        # Set the toggle device to ON state
        await self.test_device.async_turn_on()

        # Create a new manager
        new_meross_client, requires_logout = await async_get_client()
        m = None
        try:
            # Retrieve the same device with another manager
            m = MerossManager(http_client=new_meross_client)
            await m.async_init()
            await m.async_device_discovery()
            devs = m.find_devices(device_uuids=(self.test_device.uuid,))
            dev = devs[0]

            mgr_e = asyncio.Event()
            dev_e = asyncio.Event()

            # Define the coroutine for handling push notification at Manager Level
            async def manager_push_handler(push: GenericPushNotification, devices: List[BaseDevice], manager: MerossManager):
                mgr_e.set()

            # Define the coroutine for handling push notification for a specific device
            async def dev_push_handler(namespace, data, device_internal_id):
                dev_e.set()

            dev.register_push_notification_handler_coroutine(dev_push_handler)
            self.meross_manager.register_push_notification_handler_coroutine(manager_push_handler)

            await self.test_device.async_turn_off()
            aws = [asyncio.create_task(mgr_e.wait()), asyncio.create_task(dev_e.wait())]
            done, pending = await asyncio.wait(aws, timeout=5.0, return_when=asyncio.ALL_COMPLETED)
            assert(len(done) == 2 and len(pending) == 0)

            # Unregister the handlers, repeat the test and make sure none of them is triggered
            dev.unregister_push_notification_handler_coroutine(dev_push_handler)
            self.meross_manager.unregister_push_notification_handler_coroutine(manager_push_handler)
            mgr_e.clear()
            dev_e.clear()

            aws = [asyncio.create_task(mgr_e.wait()), asyncio.create_task(dev_e.wait())]
            await self.test_device.async_turn_on()
            done, pending = await asyncio.wait(aws, timeout=5.0, return_when=asyncio.ALL_COMPLETED)
            assert (len(done) == 0 and len(pending) == 2)
            for task in pending:
                task.cancel()

        finally:
            if m is not None:
                m.close()
            if requires_logout:
                await new_meross_client.async_logout()

    async def tearDownAsync(self):
        if self.requires_logout:
            await self.meross_client.async_logout()
        self.meross_manager.close()

        # Give a change to asyncio clean everything up
        await asyncio.sleep(1)


class AuthenticatedPostException(Exception):
    pass


class HttpApiError(AuthenticatedPostException):
    def __init__(self, error_code):
        self._error_code = error_code

    @property
    def error_code(self):
        return self._error_code


class BadLoginException(Exception):
    pass

class MissingMFA(BadLoginException):
    pass

class WrongMFA(BadLoginException):
    pass

class BadDomainException(Exception):
    def __init__(self, msg: str, api_domain: str, mqtt_domain:str):
        super().__init__(msg)
        self.api_domain = api_domain
        self.mqtt_domain = mqtt_domain


class UnauthorizedException(Exception):
    pass


class TokenExpiredException(Exception):
    pass


class TooManyTokensException(Exception):
    pass





from __future__ import annotations

import asyncio
import json
import logging
import time
from datetime import datetime
from typing import List, Union, Optional, Iterable, Callable, Awaitable, Dict

from meross_iot.model.constants import DEFAULT_MQTT_PORT, DEFAULT_MQTT_HOST, DEFAULT_COMMAND_TIMEOUT
from meross_iot.model.enums import OnlineStatus, Namespace
from meross_iot.model.http.device import HttpDeviceInfo
from meross_iot.model.plugin.hub import BatteryInfo
from meross_iot.utilities.network import extract_domain, extract_port

_LOGGER = logging.getLogger(__name__)


class BaseDevice(object):
    """
    A `BaseDevice` is a generic representation of a Meross device.
    Any BaseDevice is characterized by some generic information, such as user's defined
    name, type (i.e. device specific model), firmware/hardware version, a Meross internal
    identifier, a library assigned internal identifier.
    """
    _name: str = "unknown"
    _type: str = "unknown"
    _fwversion: str = "unknown"
    _hwversion: str = "unknown"
    _online: OnlineStatus = OnlineStatus.UNKNOWN
    _inner_ip: Optional[str] = None
    _mac_address: Optional[str] = None
    _mqtt_host: str = DEFAULT_MQTT_HOST
    _mqtt_port: int = DEFAULT_MQTT_PORT

    def __init__(self, device_uuid: str,
                 manager,
                 **kwargs):
        self._uuid = device_uuid
        self._manager = manager

        self._cached_http_info = None
        self._channels = []

        # Parse device info, if any
        if 'http_device_info' in kwargs:
            self._cached_http_info: HttpDeviceInfo = kwargs.get('http_device_info', {})
            self._channels = self._parse_channels(self._cached_http_info.channels)

            # Information about device
            self._name = self._cached_http_info.dev_name
            self._type = self._cached_http_info.device_type
            self._fwversion = self._cached_http_info.fmware_version
            self._hwversion = self._cached_http_info.hdware_version
            self._online = self._cached_http_info.online_status
            self._inner_ip = None

            # Domain and port
            domain = self._cached_http_info.domain
            reserved_domain = self._cached_http_info.reserved_domain

            # Prefer domain to reserved domain
            if domain is not None:
                self._mqtt_host = extract_domain(domain)
                self._mqtt_port = extract_port(domain, DEFAULT_MQTT_PORT)
            elif reserved_domain is not None:
                self._mqtt_host = extract_domain(reserved_domain)
                self._mqtt_port = extract_port(reserved_domain, DEFAULT_MQTT_PORT)
            else:
                _LOGGER.warning("No MQTT DOMAIN/RESERVED DOMAIN specified in args, assuming default value %s:%d",
                                DEFAULT_MQTT_HOST, DEFAULT_MQTT_PORT)

        if hasattr(self, "_abilities_spec"):
            self._abilities = self._abilities_spec
        else:
            self._abilities = {}
        self._push_coros = []
        self._last_full_update_ts = None

        # Set default timeout value for command execution
        self._timeout = DEFAULT_COMMAND_TIMEOUT

    @property
    def cached_http_info(self) -> Optional[HttpDeviceInfo]:
        return self._cached_http_info

    @property
    def lan_ip(self):
        return self._inner_ip

    @property
    def mac_address(self):
        return self._mac_address

    @property
    def mqtt_host(self):
        return self._mqtt_host

    @property
    def mqtt_port(self):
        return self._mqtt_port

    @property
    def abilities(self):
        return self._abilities

    @property
    def last_full_update_timestamp(self):
        return self._last_full_update_ts

    def check_full_update_done(self):
        update_done = self._last_full_update_ts is not None
        if not update_done:
            _LOGGER.error(f"Please invoke async_update() for this device ({self._name}) "
                          "before accessing its state. Failure to do so may result in inconsistent state.")
        return update_done

    def register_push_notification_handler_coroutine(self, coro: Callable[[Namespace, dict, str], Awaitable]) -> None:
        """
        Registers a coroutine so that it gets invoked whenever a push notification is
        delivered to this device or when the device state is changed.
        This allows the developer to "react" to notifications state change due to other users operating the device.
        :param coro: coroutine-function to invoke when the state changes.
        Its signature must be (namespace: Namespace, data: dict, device_internal_id: str)
        :return:
        """
        if not asyncio.iscoroutinefunction(coro):
            raise ValueError("The coro parameter must be a coroutine")
        if coro in self._push_coros:
            _LOGGER.error(f"Coroutine {coro} was already added to event handlers of this device")
            return
        self._push_coros.append(coro)

    def unregister_push_notification_handler_coroutine(self, coro: Callable[[Namespace, dict, str], Awaitable]) -> None:
        """
        Unregisters the event handler
        :param coro: coroutine-function: a function that, when invoked, returns a Coroutine object that can be awaited.
        This coroutine function should have been previously registered
        :return:
        """
        if coro in self._push_coros:
            self._push_coros.remove(coro)
        else:
            _LOGGER.error(f"Coroutine {coro} was not registered as handler for this device")

    async def _fire_push_notification_event(self, namespace: Namespace, data: dict, device_internal_id: str):
        for c in self._push_coros:
            try:
                await c(namespace=namespace, data=data, device_internal_id=device_internal_id)
            except Exception as e:
                _LOGGER.exception(f"Error occurred while firing push notification event {namespace} with data: {data}")

    @property
    def internal_id(self) -> str:
        """
        Internal ID used by this library to identify meross devices. It's basically composed by
        the Meross ID plus some prefix/suffix.
        :return:
        """
        return f"#BASE:{self._uuid}"

    @property
    def uuid(self) -> str:
        """
        Meross identifier of the device.
        :return:
        """
        return self._uuid

    @property
    def name(self) -> str:
        """
        User's defined name of the device
        :return:
        """
        return "unknown" if self._name is None else self._name

    @property
    def type(self) -> str:
        """
        Device model type
        :return:
        """
        return "unknown" if self._type is None else self._type

    @property
    def firmware_version(self) -> str:
        """
        Device firmware version. When unavailable, 'unknown' is returned
        :return:
        """
        return "unknown" if self._fwversion is None else self._fwversion

    @property
    def hardware_version(self) -> str:
        """
        Device hardware revision
        :return:
        """
        return "unknown" if self._hwversion is None else self._hwversion

    @property
    def online_status(self) -> OnlineStatus:
        """
        Current device online status
        :return:
        """
        return self._online

    @property
    def channels(self) -> List[ChannelInfo]:
        """
        List of channels exposed by this device. Multi-channel devices might expose a master
        switch at index 0.
        :return:
        """
        return self._channels

    async def update_from_http_state(self, hdevice: HttpDeviceInfo) -> BaseDevice:
        # Careful with online  status: not all the devices might expose an online mixin.
        if hdevice.uuid != self.uuid:
            raise ValueError(f"Cannot update device ({self.uuid}) with HttpDeviceInfo for device id {hdevice.uuid}")
        self._cached_http_info=hdevice
        self._cached_http_info = hdevice
        self._name = hdevice.dev_name
        self._channels = self._parse_channels(hdevice.channels)
        self._type = hdevice.device_type
        self._fwversion = hdevice.fmware_version
        self._hwversion = hdevice.hdware_version
        self._online = hdevice.online_status

        # TODO: fire some sort of events to let users see changed data?
        return self

    async def async_handle_push_notification(self, namespace: Namespace, data: dict) -> bool:
        _LOGGER.debug(f"MerossBaseDevice {self.name} handling notification {namespace}")

        # However, we want to notify any registered event handler
        await self._fire_push_notification_event(namespace=namespace, data=data, device_internal_id=self.internal_id)
        return False

    async def async_handle_update(self, namespace: Namespace, data: dict) -> bool:
        # Catch SYSTEM_ALL case and update the generic device info
        if namespace == Namespace.SYSTEM_ALL:
            # TODO: we might update name/uuid/other stuff in here...
            system = data.get('all', {}).get('system', {})
            self._inner_ip = system.get('firmware', {}).get('innerIp')
            self._mac_address = system.get('hardware', {}).get('macAddress', None)

        await self._fire_push_notification_event(namespace=namespace, data=data, device_internal_id=self.internal_id)
        self._last_full_update_ts = time.time() * 1000

        # Even though we handle the event, we return False as we did not handle the event in any way
        # rather than updating the last_full_update_ts
        return False

    async def async_update(self,
                           *args,
                           **kwargs) -> None:
        """
        Forces a full data update on the device. If your network bandwidth is limited or you are running
        this program on an embedded device, try to invoke this method only when strictly needed.
        Most of the parameters of a device are updated automatically upon push-notification received
        by the meross MQTT cloud.
        :return: None
        """
        """
        # This method should be overridden implemented by mixins and never called directly. Its main
        # objective is to call the corresponding GET ALL command, which varies in accordance with the
        # device type. For instance, wifi devices use GET System.Appliance.ALL while HUBs use a different one.
        # Implementing mixin should never call the super() implementation (as it happens
        # with _handle_update) as we want to use only an UPDATE_ALL method.
        # Howe                               ver, we want to keep it within the MerossBaseDevice so that we expose a consistent
        # interface.
        """
        pass

    def dismiss(self):
        # TODO: Should we do something here?
        pass

    @property
    def default_command_timeout(self):
        """
        Represents the default timeout that is applied to command execution against this device.
        Usually, every method allows to override this timeout via an appropriate timeout argument: that argument
        takes precedence over this default.
        """
        return self._timeout

    @default_command_timeout.setter
    def default_command_timeout(self, val: Union[float, int]):
        if val is None or val < 0:
            raise ValueError("Command execution timeout must a positive number")
        self._timeout = val

    async def _execute_command(self,
                               method: str,
                               namespace: Namespace,
                               payload: dict,
                               timeout: Optional[float] = None,
                               ) -> dict:
        if timeout is None:
            to = self.default_command_timeout
        else:
            to = timeout

        return await self._manager.async_execute_cmd(destination_device_uuid=self.uuid,
                                                     method=method,
                                                     namespace=namespace,
                                                     payload=payload,
                                                     timeout=to,
                                                     mqtt_hostname=self.mqtt_host,
                                                     mqtt_port=self.mqtt_port)

    def __repr__(self):
        basic_info = f"{self.name} ({self.type}, HW {self.hardware_version}, FW {self.firmware_version}, class: {self.__class__.__name__})"
        return basic_info

    @staticmethod
    def _parse_channels(channel_data: List) -> List[ChannelInfo]:
        res = []
        if channel_data is None:
            return res

        for i, val in enumerate(channel_data):
            name = val.get('devName', 'Main channel')
            type = val.get('type')
            master = i == 0
            res.append(ChannelInfo(index=i, name=name, channel_type=type, is_master_channel=master))

        return res

    def lookup_channel(self, channel_id_or_name: Union[int, str]):
        """
        Looks up a channel by channel id or channel name
        :param channel_id_or_name:
        :return:
        """
        res = []
        if isinstance(channel_id_or_name, str):
            res = list(filter(lambda c: c.name == channel_id_or_name, self._channels))
        elif isinstance(channel_id_or_name, int):
            res = list(filter(lambda c: c.index == channel_id_or_name, self._channels))
        if len(res) == 1:
            return res[0]
        raise ValueError(f"Could not find channel by id or name = {channel_id_or_name}")

    def support_encryption(self) -> bool:
        """
        Returns true if encryption is supported by this device
        :return:
        """
        return False

    def is_encryption_key_set(self) -> bool:
        """
        Returns whether an encryption key has been already set
        :return:
        """
        return False

    def set_encryption_key(self, *args, **kwargs):
        """
        Sets the encryption key to be used for encryption and decryption
        :param args:
        :param kwargs:
        :return:
        """
        pass

    def encrypt(self, message_data_bytes: bytes)->str:
        """
        Encrypts the message into a base64 string
        :param message_data_bytes:
        :return:
        """
        raise NotImplementedError("Encryption not supported by this device")

    def decrypt(self, encrypted_message_bytes: bytes) -> bytes:
        """
        Decrypt the message and returns the war decrypted bytes
        :param encrypted_message_bytes:
        :return:
        """
        raise NotImplementedError("Encryption not supported by this device")


class HubDevice(BaseDevice):
    # TODO: provide meaningful comment here describing what this class does
    #  Discvoery?? Bind/unbind?? Online??
    def __init__(self, device_uuid: str, manager, **kwargs):
        super().__init__(device_uuid, manager, **kwargs)
        self._sub_devices = {}

    def get_subdevices(self) -> Iterable[GenericSubDevice]:
        return self._sub_devices.values()

    def get_subdevice(self, subdevice_id: str) -> Optional[GenericSubDevice]:
        return self._sub_devices.get(subdevice_id)

    def register_subdevice(self, subdevice: GenericSubDevice) -> None:
        # If the device is already registed, skip it
        if subdevice.subdevice_id in self._sub_devices:
            _LOGGER.info(f"Subdevice {subdevice.subdevice_id} has been already registered to this HUB ({self.name})")
            return

        self._sub_devices[subdevice.subdevice_id] = subdevice


class GenericSubDevice(BaseDevice):
    _UPDATE_ALL_NAMESPACE = None

    def __init__(self, hubdevice_uuid: str, subdevice_id: str, manager, **kwargs):
        hubs = manager.find_devices(device_uuids=(hubdevice_uuid,))  # type: List[HubDevice]
        if len(hubs) < 1:
            raise ValueError("Specified hub device is not present")
        hub = hubs[0]
        super().__init__(device_uuid=hubdevice_uuid, manager=manager, domain=hub.mqtt_host, port=hub.mqtt_port,
                         **kwargs)
        self._subdevice_id = subdevice_id
        self._type = kwargs.get('subDeviceType')
        self._name = kwargs.get('subDeviceName')
        self._onoff = None
        self._mode = None
        self._temperature = None
        self._hub = hub

    async def _execute_command(self,
                               method: str,
                               namespace: Namespace,
                               payload: dict,
                               timeout: Optional[float] = None
                               ) -> dict:
        # Every command should be invoked via HUB?
        raise NotImplementedError("Subdevices should rely on Hub in order to send commands.")

    async def async_update(self,
                           timeout: Optional[float] = None,
                           *args,
                           **kwargs) -> None:
        """
        Perfoms a full device update of the device attributes.
        """
        if self._UPDATE_ALL_NAMESPACE is None:
            _LOGGER.error("GenericSubDevice does not implement any GET_ALL namespace. Update won't be performed.")
            pass

        # When dealing with hubs, we need to "intercept" the UPDATE()
        await super().async_update(*args, **kwargs)

        # When issuing an update-all command to the hub,
        # we need to query all sub-devices.
        result = await self._hub._execute_command(method="GET",
                                                  namespace=self._UPDATE_ALL_NAMESPACE,
                                                  payload={'all': [{'id': self.subdevice_id}]},
                                                  timeout=timeout)
        subdevices_states = result.get('all')
        for subdev_state in subdevices_states:
            subdev_id = subdev_state.get('id')

            if subdev_id != self.subdevice_id:
                continue
            await self.async_handle_subdevice_notification(namespace=self._UPDATE_ALL_NAMESPACE, data=subdev_state)
            break

    async def async_get_battery_life(self,
                                     timeout: Optional[float] = None,
                                     *args,
                                     **kwargs) -> BatteryInfo:
        """
        Polls the HUB/DEVICE to get its current battery status.
        :return:
        """
        data = await self._hub._execute_command(method='GET',
                                                namespace=Namespace.HUB_BATTERY,
                                                payload={'battery': [{'id': self.subdevice_id}]},
                                                timeout=timeout)
        battery_life_perc = data.get('battery', {})[0].get('value')
        timestamp = datetime.utcnow()
        return BatteryInfo(battery_charge=battery_life_perc, sample_ts=timestamp)

    async def async_handle_subdevice_notification(self, namespace: Namespace, data: dict) -> bool:
        _LOGGER.error("Unhandled/NotImplemented event handler for %s (data: %s) - Subdevice %s (hub %s)", namespace,
                      json.dumps(data), self.subdevice_id, self._hub.uuid)
        return False

    @property
    def internal_id(self) -> str:
        return f"#BASE:{self._uuid}#SUB:{self._subdevice_id}"

    @property
    def subdevice_id(self):
        return self._subdevice_id

    @property
    def online_status(self) -> OnlineStatus:
        # If the HUB device is offline, return offline
        if self._hub.online_status != OnlineStatus.ONLINE:
            return self._hub.online_status

        return self._online

    def _prepare_push_notification_data(self, data: dict, filter_accessor: str = None) -> Optional[Dict]:
        if filter_accessor is not None:
            # Operate only on relative accessor
            context = data.get(filter_accessor)
            if context is None:
                raise ValueError(
                    "Could not find accessor %s within data %s. This push notification will be ignored." % (
                        filter_accessor, str(data)))
            pertinent_notifications = filter(lambda n: n.get('id') == self.subdevice_id, context)
            next(pertinent_notifications, None)

        else:
            notification = data.copy()
            if 'id' in notification:
                if notification.get('id') != self.subdevice_id:
                    _LOGGER.error("Ignoring notification %s as it does not target "
                                  "to subdevice id %s", notification, self.subdevice_id)
                notification.pop('id')
            return notification


class ChannelInfo(object):
    def __init__(self, index: int, name: str = None, channel_type: str = None, is_master_channel: bool = False):
        self._index = index
        self._name = name
        self._type = channel_type
        self._master = is_master_channel

    @property
    def index(self) -> int:
        return self._index

    @property
    def is_usb(self) -> bool:
        return self._type == 'USB'

    @property
    def is_master_channel(self) -> bool:
        return self._master

    @property
    def name(self) -> str:
        return self._name



import logging
from multiprocessing import Process

import socks
from aiohttp import web
from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop

from meross_iot.controller.device import BaseDevice
from meross_iot.controller.mixins.toggle import ToggleXMixin
from meross_iot.manager import MerossManager
from meross_iot.model.enums import OnlineStatus, Namespace
from tests import async_get_client
import os
from proxy import proxy

PROXY_PORT = 9001
_LOGGER = logging.getLogger(__name__)


if os.name == 'nt':
    import asyncio

    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
else:
    import asyncio


class ProxyManager:
    def __init__(self, port):
        self._port = port
        self._process = None
        self._started = False

    def start(self):
        if not self._started:
            self._process = Process(target=proxy.main, args=(['--port', str(self._port), '--hostname', '127.0.0.1', '--num-workers', '2'],))
            self._process.start()
            self._started = True

    def stop(self):
        if self._started:
            self._process.terminate()
            self._process.join(5)
            if self._process.exitcode is None:
                self._process.kill()
            self._process = None
            self._started = False


class TestDisconnection(AioHTTPTestCase):
    async def get_application(self):
        return web.Application()

    async def setUpAsync(self):
        # Setup the proxy
        self._proxy = ProxyManager(port=PROXY_PORT)
        self._proxy.start()
        _LOGGER.info("Proxy started")

        # Wait some time before next test-burst
        _LOGGER.info("Sleeping...")
        await asyncio.sleep(10)
        _LOGGER.info("Allocating and configuring MerossManager to use the proxy")
        self.meross_client, self.requires_logout = await async_get_client()
        self.meross_client.set_http_proxy(f"http://localhost:{PROXY_PORT}")

        self.meross_manager = MerossManager(http_client=self.meross_client)
        self.meross_manager.set_proxy(proxy_type=socks.HTTP, proxy_addr="localhost", proxy_port=PROXY_PORT)

        await self.meross_manager.async_init()
        _LOGGER.info("Manager started")

        await self.meross_manager.async_device_discovery()
        self.test_devices = self.meross_manager.find_devices(device_class=ToggleXMixin,
                                                             online_status=OnlineStatus.ONLINE)
        _LOGGER.info("Discovery ended. Test devices: %s", self.test_devices)

    @unittest_run_loop
    async def test_disconnect(self):
        if len(self.test_devices) < 1:
            self.skipTest("No ToggleX device has been found to run this test on.")

        # Select the first device
        dev = self.test_devices[0]  # type: BaseDevice
        print(f"Testing device {dev.name}")
        _LOGGER.info("Selected device %s for disconnection testing", dev)
        await dev.async_update()
        _LOGGER.info("Device state updated")

        # Register a connection reset event handler
        disconnection_event = asyncio.Event()
        connection_event = asyncio.Event()

        async def evt_coro(namespace: Namespace, data: dict, device_internal_id: str, *args, **kwargs):
            if namespace == Namespace.SYSTEM_ONLINE:
                status = OnlineStatus(int(data.get('online').get('status')))
                if not disconnection_event.is_set() and status == OnlineStatus.UNKNOWN:
                    _LOGGER.info("Disconnection event caught")
                    disconnection_event.set()
                elif not connection_event.is_set() and status == OnlineStatus.ONLINE:
                    _LOGGER.info("Connection event caught")
                    connection_event.set()

        dev.register_push_notification_handler_coroutine(evt_coro)
        _LOGGER.info("Stopping proxy")
        self._proxy.stop()

        try:
            _LOGGER.info("Waiting for disconnection event...")
            await asyncio.wait_for(disconnection_event.wait(), 90.0)
        except asyncio.TimeoutError:
            self.fail("Disconnection was not triggered.")

        _LOGGER.info("Disconnection test OK!")

        # Ok, disconnection was triggered. Let's now re-enable the proxy and wait for devices coming online again...
        _LOGGER.info("Starting proxy again to trigger reconnection...")
        self._proxy.start()
        _LOGGER.info("Proxy started. Waiting a bit.")
        await asyncio.sleep(10)
        # Let's wait for connection to be re-established
        try:
            _LOGGER.info("Waiting for connection event...")
            await asyncio.wait_for(connection_event.wait(), 90.0)
        except asyncio.TimeoutError:
            self.fail("Connection was not triggered.")

        _LOGGER.info("Reconnection test OK!")

    async def tearDownAsync(self):
        if self.requires_logout:
            await self.meross_client.async_logout()

        self._proxy.stop()
        self.meross_manager.close()

        # Give a change to asyncio clean everything up
        await asyncio.sleep(1)



class UnconnectedError(Exception):
    pass


class CommandTimeoutError(Exception):
    def __init__(self, message: str, target_device_uuid: str, timeout: float):
        self.message = message
        self.target_device_uuid = target_device_uuid
        self.timeout = timeout


class MqttError(Exception):
    def __init__(self, message=None):
        super().__init__()
        self._message = message


class CommandError(Exception):
    def __init__(self, error_payload: dict):
        super().__init__()
        self.error_payload = error_payload


class UnknownDeviceType(Exception):
    pass


import os

from aiohttp import web
from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop

from meross_iot.controller.mixins.garage import GarageOpenerMixin
from meross_iot.controller.mixins.toggle import ToggleXMixin
from meross_iot.http_api import MerossHttpClient
from meross_iot.manager import MerossManager
from meross_iot.model.enums import OnlineStatus
from tests import async_get_client

if os.name == 'nt':
    import asyncio
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
else:
    import asyncio


class TestToggleX(AioHTTPTestCase):
    async def get_application(self):
        return web.Application()

    async def setUpAsync(self):
        # Wait some time before next test-burst
        await asyncio.sleep(10)
        self.meross_client, self.requires_logout = await async_get_client()

        # Look for a device to be used for this test
        self.meross_manager = MerossManager(http_client=self.meross_client)
        await self.meross_manager.async_init()
        devices = await self.meross_manager.async_device_discovery()
        self.toggle_devices = self.meross_manager.find_devices(device_class=ToggleXMixin, online_status=OnlineStatus.ONLINE)
        # Filter away garage openers: they are tested independently.
        self.toggle_devices = list(filter(lambda x: not isinstance(x, GarageOpenerMixin), self.toggle_devices))

        if len(self.toggle_devices) < 1:
            self.test_device = None
        else:
            self.test_device = self.toggle_devices[0]

    @unittest_run_loop
    async def test_toggle_local_state(self):
        if self.test_device is None:
            self.skipTest("No ToggleX device has been found to run this test on.")
        print(f"Testing device {self.test_device.name}")
        # Turn off device to start from a clean state
        r = await self.test_device.async_turn_off()

        # Turn on the device
        r = await self.test_device.async_turn_on()
        self.assertTrue(self.test_device.is_on())

        # Turn off the device
        await asyncio.sleep(1)
        r = await self.test_device.async_turn_off()
        self.assertFalse(self.test_device.is_on())

    @unittest_run_loop
    async def test_toggle_multi_channel(self):
        # Search for a device with multiple channels
        multi_channel_devices = list(filter(lambda d: len(d.channels) > 1, self.toggle_devices))
        if len(multi_channel_devices) < 1:
            self.skipTest("Could not find any online device supporting more than 1 channel")

        # Toggle non master switches
        d = multi_channel_devices[0]
        print(f"Testing device {d.name}")
        for c in d.channels:
            if c.is_master_channel:
                continue
            await d.async_turn_on(channel=c.index)
            self.assertEqual(d.is_on(channel=c.index), True)
            await asyncio.sleep(1)
            await d.async_turn_off(channel=c.index)
            self.assertEqual(d.is_on(channel=c.index), False)

    @unittest_run_loop
    async def test_toggle_master_switch(self):
        # Search for a device with multiple channels
        multi_channel_devices = list(filter(lambda d: len(d.channels) > 1, self.toggle_devices))
        if len(multi_channel_devices) < 1:
            self.skipTest("Could not find any online device supporting more than 1 channel")

        # Turn on non-master switches
        d = multi_channel_devices[0]
        print(f"Testing device {d.name}")
        master = None
        for c in d.channels:
            if c.is_master_channel:
                master = c
                continue
            await d.async_turn_on(channel=c.index)
            self.assertEqual(d.is_on(channel=c.index), True)

        await asyncio.sleep(1)

        # Turn-off master switch
        self.assertIsNotNone(master)
        await d.async_turn_off(channel=master.index)

        # Give some time to the library to get the PUSH notification
        # Then make sure that the master switch has turned off all the available switches.
        await asyncio.sleep(2)
        for c in d.channels:
            self.assertEqual(d.is_on(channel=c.index), False)

    @unittest_run_loop
    async def test_usb_switches(self):
        # Search for a device with usb channel
        usb_dev = None
        usb_channel = None
        for d in self.toggle_devices:
            for c in d.channels:
                if c.is_usb:
                    usb_dev = d
                    usb_channel = c
                    break
            if usb_dev is not None:
                break

        if usb_dev is None:
            self.skipTest("Could not find any device with an usb channel")
        print(f"Testing device {usb_dev.name}")
        # Turn the channel off
        await usb_dev.async_turn_off(channel=usb_channel.index)
        self.assertFalse(usb_dev.is_on(channel=usb_channel.index))
        await asyncio.sleep(1)
        await usb_dev.async_turn_on(channel=usb_channel.index)
        self.assertTrue(usb_dev.is_on(channel=usb_channel.index))

    @unittest_run_loop
    async def test_toggle_push_notification(self):
        if self.test_device is None:
            self.skipTest("No ToggleX device has been found to run this test on.")
        print(f"Testing device {self.test_device.name}")
        # Create a new manager
        new_meross_client, requires_logout = await async_get_client()
        m = None
        try:
            # Retrieve the same device with another manager
            m = MerossManager(http_client=new_meross_client)
            await m.async_init()
            await m.async_device_discovery()
            devs = m.find_devices(device_uuids=(self.test_device.uuid,))
            dev = devs[0]

            # Turn off device to start from a clean state
            r = await self.test_device.async_turn_off()
            await asyncio.sleep(2)

            # Turn on the device
            r = await self.test_device.async_turn_on()
            # Wait a bit and make sure the other manager received the push notification
            await asyncio.sleep(2)
            self.assertTrue(self.test_device.is_on())
            self.assertTrue(dev.is_on())

            # Turn off the device
            await asyncio.sleep(1)
            r = await self.test_device.async_turn_off()
            # Wait a bit and make sure the other manager received the push notification
            await asyncio.sleep(2)
            self.assertFalse(self.test_device.is_on())
            self.assertFalse(dev.is_on())

        finally:
            if m is not None:
                m.close()
            if requires_logout:
                await new_meross_client.async_logout()

    async def tearDownAsync(self):
        if self.requires_logout:
            await self.meross_client.async_logout()
        self.meross_manager.close()

        # Give a change to asyncio clean everything up
        await asyncio.sleep(1)


import logging
from datetime import datetime
from typing import Optional

from meross_iot.model.enums import Namespace

_LOGGER = logging.getLogger(__name__)

_DATE_FORMAT = '%Y-%m-%d'


class SystemRuntimeMixin(object):
    _execute_command: callable

    def __init__(self, device_uuid: str,
                 manager,
                 **kwargs):
        super().__init__(device_uuid=device_uuid, manager=manager, **kwargs)
        self._runtime_info = {}

    async def async_update_runtime_info(self, timeout: Optional[float] = None, *args, **kwargs) -> dict:
        """
        Polls the device to gather the latest runtime information for this device.
        Note that the returned value might vary with the time as Meross could add/remove/change runtime information
        in the future.

        :return: a `dict` object containing the runtime information provided by the Meross device
        """
        result = await self._execute_command(method="GET",
                                             namespace=Namespace.SYSTEM_RUNTIME,
                                             payload={},
                                             timeout=timeout)
        data = result.get('runtime')
        self._runtime_info = data
        return data

    @property
    def cached_system_runtime_info(self) -> Optional[dict]:
        """
        Returns the latest cached runtime info. If you want a fresh value, consider using the
        `update_runtime_info` method instead.
        """
        return self._runtime_info

    async def async_update(self,
                           *args,
                           **kwargs) -> None:
        # call the superclass implementation first
        await super().async_update(*args, **kwargs)
        await self.async_update_runtime_info()



import logging
from datetime import datetime
from typing import Optional

from meross_iot.model.enums import Namespace
from meross_iot.model.plugin.power import PowerInfo

_LOGGER = logging.getLogger(__name__)


_DATE_FORMAT = '%Y-%m-%d'


class ElectricityMixin(object):
    _execute_command: callable

    def __init__(self, device_uuid: str,
                 manager,
                 **kwargs):
        super().__init__(device_uuid=device_uuid, manager=manager, **kwargs)

        # We'll hold a dictionary of lastest samples, one per channel
        self.__channel_cached_samples = {}

    async def async_get_instant_metrics(self,
                                        channel=0,
                                        timeout: Optional[float] = None,
                                        *args, **kwargs) -> PowerInfo:
        """
        Polls the device to gather the instant power consumption for this device.
        Please note that current/voltage combination may not be accurate as power is.
        So, refer to power attribute rather than calculate it as Voltage * Current.
        Avoid flooding the device by calling this methods so often. Instead. you should rely on the cached value
        offered by get_last_sample(): if it's None or if the sample_timestamp of the offered value is not recent
        enough, then you should call this method to refresh it.

        :param channel: channel where to read metrics from. Defaults to 0

        :return: a `PowerInfo` object describing the current measure data
        """
        result = await self._execute_command(method="GET",
                                             namespace=Namespace.CONTROL_ELECTRICITY,
                                             payload={'channel': channel},
                                             timeout=timeout)
        data = result.get('electricity')

        # For some reason, most of the Meross device report accurate instant power, but inaccurate voltage/current.
        current = float(data.get('current'))/1000
        voltage = float(data.get('voltage')) / 10
        power = float(data.get('power')) / 1000

        result = PowerInfo(current_ampere=current, voltage_volts=voltage, power_watts=power, sample_timestamp=datetime.utcnow())
        self.__channel_cached_samples[channel] = result
        return result

    def get_last_sample(self, channel=0, *args, **kwargs) -> Optional[PowerInfo]:
        """
        Returns the previously cached value for the sensed power information (if any).

        :param channel: The channel to gather info from

        :return:
        """
        return self.__channel_cached_samples.get(channel)
